#!/usr/bin/env python


import random
import textwrap as tw
import os.path  as op
import contextlib as ctxlib

from unittest import mock

import numpy as np
import pandas as pd

import funpack.hierarchy as hierarchy

from . import (tempdir, gen_DataTable)


def test_getHierarchyFilePath():

    dt = gen_DataTable([np.random.randint(1, 10, 10)])

    dt.vartable.loc[1, 'DataCoding'] = 123

    exp = op.join(op.dirname(hierarchy.__file__),
                  'data', 'hierarchy', 'coding123.tsv')
    assert hierarchy.getHierarchyFilePath(dt, 1) == exp


@ctxlib.contextmanager
def patchDataType(dtype):
    with mock.patch('funpack.hierarchy.dataCodingType', return_value=dtype):
        yield


def test_loadHierarchyFile():

    hier = tw.dedent("""
    coding	meaning	node_id	parent_id
    30	meaning 30	3	1
    40	meaning 40	4	1
    50	meaning 50	5	4
    20	meaning 20	2	0
    60	meaning 60	6	2
    10	meaning 10	1	0
    """).strip()

    with tempdir(), patchDataType(int):

        with open('coding123.tsv', 'wt') as f:
            f.write(hier)

        h = hierarchy.loadHierarchyFile('coding123.tsv')

        assert h.parents(10) == []
        assert h.parents(20) == []
        assert h.parents(30) == [10]
        assert h.parents(40) == [10]
        assert h.parents(50) == [40, 10]
        assert h.parents(60) == [20]



def test_codeToNumeric():

    icd10code = hierarchy.HIERARCHY_DATA_NAMES['icd10']
    icd10hier = hierarchy.getHierarchyFilePath(coding=icd10code)
    icd10hier = hierarchy.loadHierarchyFile(icd10hier)
    codes     = list(random.sample(icd10hier.codings, 20)) + ['badcode']
    exp       = [icd10hier.index(c) for c in codes[:-1]] + [np.nan]

    conv = [hierarchy.codeToNumeric(c, coding=icd10code) for c in codes]

    exp  = np.array(exp)
    conv = np.array(conv)

    expna  = np.isnan(exp)
    convna = np.isnan(conv)

    assert np.all(     expna  ==       convna)
    assert np.all(exp[~expna] == conv[~convna])

    ncodes = [(code, ncode)
              for code, ncode
              in zip(codes, conv)
              if not pd.isna(ncode)]

    codes, ncodes = zip(*ncodes)
    codes = [c.lower() for c in codes]

    assert codes == [hierarchy.numericToCode(c, coding=icd10code).lower()
                     for c in ncodes]


def test_Hierarchy():
    with tempdir():
        data = tw.dedent("""
        coding\tmeaning\tnode_id\tparent_id
        a\ta desc\t5\t0
        b\tb desc\t1\t5
        c\tc desc\t3\t5
        d\td desc\t4\t3
        e\te desc\t2\t1
        """)

        with open('coding345.tsv', 'wt') as f:
            f.write(data)

        with patchDataType(str):
            h = hierarchy.loadHierarchyFile('coding345.tsv')

        assert h.index('a') == 5
        assert h.index('b') == 1
        assert h.index('c') == 3
        assert h.index('d') == 4
        assert h.index('e') == 2
        assert h.coding(1)  == 'b'
        assert h.coding(2)  == 'e'
        assert h.coding(3)  == 'c'
        assert h.coding(4)  == 'd'
        assert h.coding(5)  == 'a'

        assert h.parents(  'a') == []
        assert h.parents(  'b') == ['a']
        assert h.parents(  'c') == ['a']
        assert h.parents(  'd') == ['c', 'a']
        assert h.parents(  'e') == ['b', 'a']
        assert h.parentIDs( 5)  == []
        assert h.parentIDs( 1)  == [5]
        assert h.parentIDs( 3)  == [5]
        assert h.parentIDs( 4)  == [3, 5]
        assert h.parentIDs( 2)  == [1, 5]

        assert h.description('a') == 'a desc'
        assert h.description('b') == 'b desc'
        assert h.description('c') == 'c desc'
        assert h.description('d') == 'd desc'
        assert h.description('e') == 'e desc'

        h.set('a', 'meta', 'aa')
        h.set('b', 'meta', 'bb')
        h.set('c', 'meta', 'cc')
        h.set('d', 'meta', 'dd')
        h.set('e', 'meta', 'ee')

        assert h.get('a', 'meta') == 'aa'
        assert h.get('b', 'meta') == 'bb'
        assert h.get('c', 'meta') == 'cc'
        assert h.get('d', 'meta') == 'dd'
        assert h.get('e', 'meta') == 'ee'


def test_Hierarchy_non_sequential_ids():
    with tempdir():
        data = tw.dedent("""
        coding\tmeaning\tnode_id\tparent_id
        100\t100 desc\t550\t0
        200\t200 desc\t150\t550
        300\t300 desc\t350\t550
        400\t400 desc\t450\t350
        500\t500 desc\t250\t150
        """)

        with open('coding456.tsv', 'wt') as f:
            f.write(data)

        with patchDataType(str):
            h = hierarchy.loadHierarchyFile('coding456.tsv')

        assert h.index('100') == 550
        assert h.index('200') == 150
        assert h.index('300') == 350
        assert h.index('400') == 450
        assert h.index('500') == 250
        assert h.coding(150)  == '200'
        assert h.coding(250)  == '500'
        assert h.coding(350)  == '300'
        assert h.coding(450)  == '400'
        assert h.coding(550)  == '100'

        assert h.parents('100') == []
        assert h.parents('200') == ['100']
        assert h.parents('300') == ['100']
        assert h.parents('400') == ['300', '100']
        assert h.parents('500') == ['200', '100']
        assert h.parentIDs(550) == []
        assert h.parentIDs(150) == [550]
        assert h.parentIDs(350) == [550]
        assert h.parentIDs(450) == [350, 550]
        assert h.parentIDs(250) == [150, 550]

        assert h.description('100') == '100 desc'
        assert h.description('200') == '200 desc'
        assert h.description('300') == '300 desc'
        assert h.description('400') == '400 desc'
        assert h.description('500') == '500 desc'

        h.set('100', 'meta', 'aa')
        h.set('200', 'meta', 'bb')
        h.set('300', 'meta', 'cc')
        h.set('400', 'meta', 'dd')
        h.set('500', 'meta', 'ee')

        assert h.get('100', 'meta') == 'aa'
        assert h.get('200', 'meta') == 'bb'
        assert h.get('300', 'meta') == 'cc'
        assert h.get('400', 'meta') == 'dd'
        assert h.get('500', 'meta') == 'ee'



#!/usr/bin/env python
#
# config.py - Parses command line arguments and configuration files.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module contains functions for parsing ``funpack`` command line
arguments and configuration files.

A ``funpack`` configuration file is simply a plain-text file which contains
command-line options, sans-hyphens.
"""


import os.path         as op
import functools       as ft
import itertools       as it
import multiprocessing as mp
import                    sys
import                    shlex
import                    logging
import                    argparse
import                    collections

import funpack
import funpack.util           as util
import funpack.custom         as custom
import funpack.importing      as importing
import funpack.exporting_hdf5 as exporting_hdf5


log = logging.getLogger(__name__)


VERSION                    = funpack.version()
DEFAULT_MERGE_AXIS         = importing.MERGE_AXIS
DEFAULT_MERGE_STRATEGY     = importing.MERGE_STRATEGY
AVAILABLE_MERGE_AXES       = importing.MERGE_AXIS_OPTIONS
AVAILABLE_MERGE_STRATEGIES = importing.MERGE_STRATEGY_OPTIONS
DEFAULT_HDF5_KEY           = exporting_hdf5.HDF5_KEY
DEFAULT_HDF5_STYLE         = exporting_hdf5.HDF5_STYLE
AVAILABLE_HDF5_STYLES      = exporting_hdf5.HDF5_STYLES
CLI_ARGUMENTS              = collections.OrderedDict((

    ('Input/output files', [
        ((       'outfile',),          {}),
        ((       'infile',),           {'nargs'   : '+'}),
        (('e',   'encoding'),          {'action'  : 'append'}),
        (('l',   'loader'),            {'nargs'   : 2,
                                        'metavar' : ('FILE', 'LOADER'),
                                        'action'  : 'append'}),
        (('i',   'index'),             {'nargs'   : 2,
                                        'metavar' : ('FILE', 'INDEX'),
                                        'action'  : 'append'}),
        (('ma',  'merge_axis'),        {'choices' : AVAILABLE_MERGE_AXES,
                                        'default' : DEFAULT_MERGE_AXIS}),
        (('ms',  'merge_strategy'),    {'choices' : AVAILABLE_MERGE_STRATEGIES,
                                        'default' : DEFAULT_MERGE_STRATEGY}),
        (('rm',  'remove_duplicates'), {'action'  : 'store_true'}),
        (('rd',  'rename_duplicates'), {'action'  : 'store_true'}),
        (('cfg', 'config_file'),       {'action'  : 'append',
                                        'type'    : util.findConfigFile}),
        (('vf',  'variable_file'),     {'action'  : 'append',
                                        'type'    : util.findTableFile}),
        (('df',  'datacoding_file'),   {'action'  : 'append',
                                        'type'    : util.findTableFile}),
        (('tf',  'type_file'),         {'type'    : util.findTableFile}),
        (('pf',  'processing_file'),   {'type'    : util.findTableFile}),
        (('cf',  'category_file'),     {'type'    : util.findTableFile})]),

    ('Import options', [
        (('s',  'subject'),          {'action' : 'append'}),
        (('v',  'variable'),         {'action' : 'append'}),
        (('co', 'column'),           {'action' : 'append'}),
        (('c',  'category'),         {'action' : 'append'}),
        (('vi', 'visit'),            {'action' : 'append'}),
        (('ex', 'exclude'),          {'action' : 'append'}),
        (('ev', 'exclude_variable'), {'action' : 'append'}),
        (('ec', 'exclude_category'), {'action' : 'append'}),
        (('aa', 'add_aux_vars'),     {'action' : 'store_true'}),
        (('iv', 'index_visits'),     {'action' : 'store_true'}),
        (('tt', 'trust_types'),      {'action' : 'store_true'})]),

    ('Cleaning options', [
        (('sn',  'skip_insertna'),      {'action'  : 'store_true'}),
        (('scv', 'skip_childvalues'),   {'action'  : 'store_true'}),
        (('scf', 'skip_clean_funcs'),   {'action'  : 'store_true'}),
        (('sr',  'skip_recoding'),      {'action'  : 'store_true'}),
        (('nv',  'na_values'),          {'nargs'   : 2,
                                         'action'  : 'append',
                                         'metavar' : ('VID', 'NAVALUES')}),
        (('re',  'recoding'),           {'nargs'   : 3,
                                         'action'  : 'append',
                                         'metavar' : ('VID',
                                                      'RAWLEVELS',
                                                      'NEWLEVELS')}),
        (('cv',  'child_values'),       {'nargs'   : 3,
                                         'action'  : 'append',
                                         'metavar' : ('VID',
                                                      'EXPRS',
                                                      'VALUES')}),
        (('cl',  'clean'),              {'nargs'   : 2,
                                         'action'  : 'append',
                                         'metavar' : ('VID', 'EXPR')}),
        (('tc',  'type_clean'),         {'nargs'   : 2,
                                         'action'  : 'append',
                                         'metavar' : ('TYPE', 'EXPR')}),
        (('gc',  'global_clean'),       {'metavar' : 'EXPR'})]),

    ('Processing options', [
        (('sp',  'skip_processing'), {'action'  : 'store_true'}),
        (('ppr', 'prepend_process'), {'action'  : 'append',
                                      'nargs'   : 2,
                                      'metavar' : ('VARS', 'PROCS')}),
        (('apr', 'append_process'),  {'action'  : 'append',
                                      'nargs'   : 2,
                                      'metavar' : ('VARS', 'PROCS')})]),

    ('Export options', [
        (('io',  'ids_only'),              {'action'  : 'store_true'}),
        (('f',   'format'),                {}),
        (('edf', 'date_format'),           {'default' : 'default'}),
        (('etf', 'time_format'),           {'default' : 'default'}),
        (('evf', 'var_format'),            {'nargs'   : 2,
                                            'metavar' : ('VID', 'FORMATTER'),
                                            'action'  : 'append'}),
        (('esn', 'suppress_non_numerics'), {'action'  : 'store_true'}),
        (('nr',  'num_rows'),              {'type'    : int})]),

    ('TSV/CSV export options', [
        (('ts',  'tsv_sep'),            {}),
        (('en',  'escape_newlines'),    {'action'  : 'store_true'}),
        (('tm',  'tsv_missing_values'), {'default' : ''})]),

    ('HDF5 export options', [
        (('hk', 'hdf5_key'),   {'default' : DEFAULT_HDF5_KEY}),
        (('hs', 'hdf5_style'), {'default' : DEFAULT_HDF5_STYLE,
                                'choices' : AVAILABLE_HDF5_STYLES})]),

    ('Auxillary output file options', [
        (('wl',  'write_log'),          {'action' : 'store_true'}),
        (('wnn', 'write_non_numerics'), {'action' : 'store_true'}),
        (('wu',  'write_unknown_vars'), {'action' : 'store_true'}),
        (('wim', 'write_icd10_map'),    {'action' : 'store_true'}),
        (('wde', 'write_description'),  {'action' : 'store_true'}),
        (('ws',  'write_summary'),      {'action' : 'store_true'}),
        (('lf',  'log_file'),           {}),
        (('nnf', 'non_numerics_file'),  {}),
        (('uf',  'unknown_vars_file'),  {}),
        (('imf', 'icd10_map_file'),     {}),
        (('def', 'description_file'),   {}),
        (('sf',  'summary_file'),       {})]),

    ('Miscellaneous options', [
        (('V',  'version'),      {'action'  : 'version',
                                  'version' :
                                  '%(prog)s {}'.format(VERSION)}),
        (('dn', 'drop_na_rows'), {'action' : 'store_true'}),
        (('d',  'dry_run'),      {'action' : 'store_true'}),
        (('nb', 'no_builtins'),  {'action' : 'store_true'}),
        (('nj', 'num_jobs'),     {'type'    : int,
                                  'default' : 1}),
        (('p',  'plugin_file'),  {'action'  : 'append',
                                  'metavar' : 'FILE',
                                  'type'    : util.findPluginFile}),
        (('ow', 'overwrite'),    {'action'  : 'store_true'}),
        (('n',  'noisy'),        {'action'  : 'count'}),
        (('q',  'quiet'),        {'action'  : 'store_true'}),
        (('eh', 'echo'),         {'action'  : 'append'})])))


CLI_DESCRIPTIONS = {

    'Inputs' :
    'The --variable_file and --datacoding_file options can be used multiple\n'
    'times - all provided files will be merged into a single table using the\n'
    'variable/data coding IDs.',

    'Export options' :
    'Non-numeric columns are exported to the main output file by default,\n'
    'but you can control this behaviour using one of the following options:\n'
    '\n'
    ' - The --suppress_non_numerics option tells FUNPACK to only save\n'
    '   numeric columns to the main output file.\n'
    ' - The --write_non_numerics option (described in the "Auxillary output\n'
    '   file options" section) tells FUNPACK to save non-numeric columns to\n'
    '   a separate output file.\n'
    '\n'
    'Note that the --suppress_non_numerics option is independent from the\n'
    '--write_non_numerics option - if you want to save non-numeric columns\n'
    'to a separate file, instead of to the main file, you must use both\n'
    'options together.',

    'Auxillary output file options' :
    'If the --write_log option is used, a default name, based on the main\n'
    'output file name, will be given to the log file. Alternatively, the\n'
    '--log_file option can be used with a specific name to use for the log\n'
    'file. This logic also applies to the other auxillary output files.\n'
    '\n'
    'The --unknown_vars_file allows a file to be saved containing\n'
    'information about columns which were in the input data, but were either\n'
    'not in the variable table, or were uncategorised and did not have any\n'
    'cleaning/processing rules specified. It contains four columns - the\n'
    'column name, the originating input file, the reason the column is being\n'
    'included (either unknown or uncategorised), and whether the column was\n'
    'exported.',
}


CLI_ARGUMENT_HELP = {

    # Input options
    'outfile' : 'Location to store output data.',
    'infile'  : 'File(s) containing input data.',

    'encoding' :
    'Character encoding. A single encoding can be specified, or this option '
    'can be used multiple times, one for each input file.',

    'loader' :
    'Use custom loader for file. Can be used multiple times.',

    'index' :
    'Position of index column for file (starting from 0). Can be used '
    'multiple times. Defaults to 0. Specify multi-column indexes as '
    'comma-separated lists.',

    'merge_axis' :
    'Axis to concatenate multiple input files along (default: "{}"). '
    'Options are "subjects"/"rows"/"0" or "variables"/"columns"/'
    '"1".'.format(DEFAULT_MERGE_AXIS),

    'merge_strategy' :
    'Strategy for merging multiple input files (default: "{}"). '
    'Options are "naive", "intersection"/"inner", or "union"/'
    '"outer".'.format(DEFAULT_MERGE_STRATEGY),

    'remove_duplicates' :
    'Remove duplicate columns, only retaining the first.',

    'rename_duplicates' :
    'Rename any duplicate columns so that all columns have a unique name.',

    'config_file' :
    'File containing default command line arguments. Can be used multiple '
    'times.',

    'variable_file' :
    'File(s) containing rules for handling variables',

    'datacoding_file' :
    'File(s) containing rules for handling data codings.',

    'type_file' :
    'File containing rules for handling types.',

    'processing_file' :
    'File containing variable processing rules.',

    'category_file' :
    'File containing variable categories.',

    'trust_types' :
    'Assume that columns in the input data with a known numeric type do not '
    'contain any bad/unparseable values. Using this flag will improve import '
    'performance, but will cause funpack to crash if the input file(s) do '
    'contain bad values.',

    'subject' :
    'Subject ID, range, comma-separated list, or file containing a list of '
    'subject IDs, or variable expression, denoting subjects to include. Can '
    'be used multiple times.',

    'variable' :
    'Variable ID, range, comma-separated list, or file containing a list of '
    'variable IDs, to import. Can be used multiple times.',

    'column' :
    'Name of column to import, or file containing a list of column names. Can '
    'also be a glob-style wildcard pattern - columns with a name matching the '
    'pattern will be imported. Can be used multiple times.',

    'category' :
    'Category ID or label to import. Can be used multiple times.',

    'visit' :
    'Import this visit. Can be used multiple times. Allowable values are '
    '\'first\', \'last\', or a visit number.',

    'exclude' :
    'Subject ID, range, comma-separated list, or file containing a list of '
    'subject IDs specifying subjects to exclude. Can be used multiple times.',

    'exclude_variable' :
    'Variable ID, range, comma-separated list, or file containing a list of '
    'variable IDs, to exclude. Takes precedence over the --variable and '
    '--category options. Can be used multiple times.',

    'exclude_category' :
    'Category ID or label to exclude. Takes precedence over the --variable '
    'and --category options. Can be used multiple times.',

    'add_aux_vars' :
    'Automatically import auxillary variables which are specified in '
    'processing rules if not already selected. Note that this only affects '
    'auxillary variables - see the "take" option to the binariseCategorical '
    'processing function for an example of an auxillary variable.',

    'index_visits' :
    'If set, the data is re-arranged so that visits form part of the row '
    'indices, rather than being stored in separate columns for each variable. '
    'Only applied to variables which are labelled with instancing 2 (Biobank '
    'assessment centre visit).',

    # Clean options
    'skip_insertna' :
    'Skip NA insertion.',
    'skip_childvalues' :
    'Skip child value replacement.',
    'skip_clean_funcs' :
    'Skip cleaning functions defined in variable table.',
    'skip_recoding' :
    'Skip raw->new level recoding.',

    'na_values' :
    'Replace these values with NA (overrides any NA values specified in '
    'variable table). Can be used multiple times. Values must be specified as '
    'a comma-separated list.',

    'recoding' :
    'Recode categorical variables (overrides any raw/new level recodings '
    'specified in variable table). Can be used multiple times. Raw and new '
    'level values must be specified as comma-separated lists.',

    'child_values' :
    'Replace NA with the given values where the given expressions evaluate '
    'to true (overrides any parent/child values specified in variable table). '
    'Can be used multiple times. Parent value expressions and corresponding '
    'child values must be specified as comma-separated lists.',

    'clean' :
    'Apply cleaning function(s) to variable (overrides any cleaning '
    'specified in variable table).',

    'type_clean' :
    'Apply clean function(s) to all variables of the specified type '
    '(overrides any cleaning specified in type table).',

    'global_clean' :
    'Apply cleaning function(s) to every variable (after variable-'
    'specific cleaning specified in variable table or via --clean).',

    # Processing options
    'skip_processing' :
    'Skip processing functions defined in processing table (but not '
    'those specified via --prepend_process and --append_process).',
    'prepend_process' :
    'Apply processing function(s) before processing defined in processing '
    'table.',
    'append_process' :
    'Apply processing function(s) after processing defined in processing '
    'table.',

    # Export options
    'ids_only' :
    'Do not output any data - instead, output a plain-text file which just '
    'contains the subject IDs, one per line. When this option is used, all '
    'cleaning and processing routines will be skipped, and no other output '
    'files (with the exception of a log file) will be produced.',
    'format' :
    'Output file format (default: inferred from the output file suffix - one '
    'of "tsv", "csv", or "h5").',
    'date_format' :
    'Formatter to use for date variables (default: "default").',
    'time_format' :
    'Formatter to use for time variables (default: "default").',
    'var_format' :
    'Apply custom formatter to the specified variable.',
    'num_rows' :
    'Number of rows to write at a time. Ignored if --num_jobs is set to 1.',
    'suppress_non_numerics' :
    'Do not save non-numeric columns to the main output file.',

    # TSV/CSV export options
    'escape_newlines' :
    'Escape any newline characters in all non-numeric columns, replacing them '
    'with a literal "\n" (with the same logic applied to any other escape '
    'characters).',
    'tsv_sep' :
    'Column separator string to use in output file (default: "," for csv, '
    '"\\t" for tsv).',
    'tsv_missing_values' :
    'String to use for missing values in output file (default: empty '
    'string).' ,

    # HDF5 export options
    'hdf5_key'    :
    'Key/name to use for the HDF5 group (default: '
    '"{}").'.format(DEFAULT_HDF5_KEY),
    'hdf5_style'  :
    'HDF5 style to save as (default: "{}").'.format(DEFAULT_HDF5_STYLE),

    # aux file options
    'write_log' : 'Save log messages to file.',
    'write_non_numerics' : 'Save non-numeric columns to file.',
    'write_unknown_vars' :
    'Save list of unknown/uncategorised variables/columns to file.',
    'write_icd10_map' :
    'Save converted ICD10 code mappings to file',
    'write_description'  :
    'Save descriptions of each column to file',
    'write_summary' :
    'Save summary of cleaning applied to each column to file',

    'log_file' : 'Save log messages to file.',
    'non_numerics_file' : 'Save non-numeric columns to file.',
    'unknown_vars_file' :
    'Save list of unknown/uncategorised variables/columns to file.',
    'icd10_map_file' :
    'Save converted ICD10 code mappings to file',
    'description_file' :
    'Save descriptions of each column to file',
    'summary_file' :
    'Save summary of cleaning applied to each column to file',

    # Miscellaneous options
    'version'      : 'Print version and exit.',
    'dry_run'      : 'Print a summary of what would happen and exit.',
    'drop_na_rows' : 'Drop rows which do not contain data for any column. '
                     'will take place on both import and export.',
    'no_builtins'  : 'Do not use the built in variable or data coding tables.',
    'num_jobs'     : 'Maximum number of jobs to run in parallel. Set to 1 '
                     '(the default) to disable parallelisation.  Set to -1 '
                     'to use all available cores ({} on this '
                     'platform).'.format(mp.cpu_count()),
    'plugin_file'  : 'Path to file containing custom funpack plugins. Can be '
                     'used multiple times.',
    'overwrite'    : 'Overwrite output file if it already exists',
    'noisy'        : 'Print debug statements. Can be used up to three times.',
    'quiet'        : 'Be quiet.',
    'echo'         : 'Echo the given value when parsed. This option is '
                     'intended for use within FUNPACK configuration files, '
                     'for example to print out the configuration file '
                     'version.',
}


def makeEpilog():
    """Generates an epilog for the command line help.

    The epilog contains an overview of available plugin functions.
    """

    def genTable(pluginType):
        plugins = custom.listPlugins(pluginType)
        descs   = []

        # The first two lines of all plugin function
        # docstrings are assumed to be:
        #   - function signature
        #   - one line description
        for p in plugins:
            try:
                desc = custom.get(pluginType, p).__doc__.split('\n')
                sig  = desc[0]
                desc = desc[1]
                descs.append((sig, desc))
            except Exception:
                descs.append(('n\a', None))

        maxpluginlen = max([len(p) for p in plugins])

        lines = []
        fmt   = '  - {{:{}s}}  {{}}'.format(maxpluginlen)

        for p, d in zip(plugins, descs):
            sig, desc = d
            lines.append(fmt.format(p, sig))
            if desc is not None:
                lines.append('{}  {}'.format(' ' * maxpluginlen, desc))

        return '\n'.join(lines)

    epilog = 'Available cleaning routines:'   + '\n'   + \
             genTable('cleaner')              + '\n\n' + \
             'Available processing routines:' + '\n'   + \
             genTable('processor')            + '\n'

    return epilog


@ft.lru_cache()
def makeParser(include=None, exclude=None):
    """Creates and returns an ``argparse.ArgumentParser`` for parsing
    ``funpack`` command line arguments.

    :arg include: Configure the parser so it only includes the specified
                  arguments.

    :arg exclude: Configure the parser so it excludes the specified
                  arguments - this overrides ``include``.
    """

    parser = argparse.ArgumentParser(
        'funpack',
        allow_abbrev=False,
        epilog=makeEpilog(),
        formatter_class=argparse.RawDescriptionHelpFormatter)
    helps  = CLI_ARGUMENT_HELP
    descs  = CLI_DESCRIPTIONS

    for group, args in CLI_ARGUMENTS.items():

        desc  = descs.get(group, None)
        group = parser.add_argument_group(group, description=desc)

        for arg, kwargs in args:

            name = arg[-1]

            if exclude is not None and name     in exclude: continue
            if include is not None and name not in include: continue

            if len(arg) == 2:
                arg = ('-{}'.format(arg[0]), '--{}'.format(arg[1]))

            group.add_argument(*arg, help=helps[name], **kwargs)

    return parser


def sanitiseArgs(argv):
    """Sanitises command-line arguments to work around a bug in ``argparse``.

    The ``argparse`` module does not work with non-numeric optional argument
    values that begin with a hyphen, as it thinks that they are argument names
    (see https://bugs.python.org/issue9334).

    This function searches for relevant optional argument names, and prepends
    a space to their values to make sure that ``argparse`` doesn't fall over.

    :arg argv: Command-line arguments
    :returns:  Sanitised command-line arguments
    """
    argv       = list(argv)
    toSanitise = [( '-nv',           [1]),
                  ('--na_values',    [1]),
                  ( '-re',           [1, 2]),
                  ('--recoding',     [1, 2]),
                  ( '-cv',           [2]),
                  ('--child_values', [2])]

    for arg, idxs in toSanitise:

        try:               argidx = argv.index(arg)
        except ValueError: continue

        for i in idxs:

            i   = i + 1 + argidx
            val = argv[i]

            if val.startswith('-'):
                argv[i] = ' ' + val

    return argv


def loadConfigFile(cfgfile, namespace=None):
    """Loads arguments from the given configuration file, returning an
    ``argparse.Namespace`` object.

    :arg cfgfile:   Path to configuration file
    :arg namespace: Existing ``argparse.Namespace`` object to merge
                    arguments into.
    """

    argv = []

    log.debug('Loading arguments from configuration file %s', cfgfile)

    with open(cfgfile, 'rt') as f:
        lines = [line.rsplit('#')[0] for line in f.readlines()]
        lines = [line.strip()        for line in lines]
        lines = [line                for line in lines if line != '']

    for line in lines:
        words        = list(shlex.split(line))
        name, values = words[0], words[1:]

        argv.append('--{}'.format(name))
        argv.extend(values)

    argv      = sanitiseArgs(argv)
    parser    = makeParser(exclude=('outfile', 'infile'))
    namespace = parser.parse_args(argv, namespace)

    # a config file may "include" other
    # config files - we parse them recursively,
    # and merge the arguments in load order.
    cfgfiles = namespace.config_file
    if cfgfiles is not None and len(cfgfiles) > 0:
        for cf in cfgfiles:
            namespace.config_file = None
            namespace             = loadConfigFile(cf, namespace)

    return namespace


def parseArgsWithConfigFile(argv=None):
    """Checks the command line arguments to see if one or more configuration
    files has been specified. If so, loads the arguments in the configuration
    file(s), and then parses the rest of the command line arguments.

    :returns: see :func:`parseArgs`.
    """

    if argv is None:
        argv = sys.argv[1:]

    argv      = sanitiseArgs(argv)
    cfgfiles  = None
    namespace = None

    if '-cfg' in argv or '--config_file' in argv:
        cfgparser = makeParser('config_file')
        cfgfiles  = cfgparser.parse_known_args(argv)[0].config_file

    if cfgfiles is not None:
        for cf in cfgfiles:
            namespace = loadConfigFile(cf, namespace)

    return parseArgs(argv, namespace)


def parseArgs(argv=None, namespace=None):
    """Parses ``funpack`` command line arguments.

    :arg argv:      List of arguments to parse.

    :arg namespace: Existing ``argparse.Namespace`` - if not provided, an
                    empty one will be created.

    :returns:       A tuple containing:

                     - an ``argparse.Namespace`` object containing the parsed
                       command-line arguments.
                     - A list of the original arguments that were parsed.
    """

    if argv is None:
        argv = sys.argv[1:]

    argv = sanitiseArgs(argv)
    args = makeParser().parse_args(argv, namespace)

    # error if output file exists
    if (not args.dry_run) and (not args.overwrite) and op.exists(args.outfile):
        print('Output file already exists, and --overwrite was not '
              'specified!')
        sys.exit(1)

    # -1 implies max-parallel
    if   args.num_jobs <= -1: args.num_jobs = mp.cpu_count()
    elif args.num_jobs ==  0: args.num_jobs = 1

    if args.noisy is None: args.noisy = 0
    if args.quiet:         args.noisy = 0

    if args.echo is None:
        args.echo = []

    _prepareInputAndOutputFiles(        args)
    _prepareAuxFileNames(               args)
    _prepareSubjectAndVariableSelectors(args)
    _prepareCategorySelectors(          args)
    _prepareColumnSelectors(            args)
    _prepareVisitSelectors(             args)
    _prepareCleaningSelectors(          args)
    _prepareIDsOnly(                    args)

    return args, argv


def _prepareInputAndOutputFiles(args):
    """Sub-function of :func:`parseArgs`. Prepares arguments related to
    the primary input and output files.

    :arg args: ``argparse.Namespace`` object.
    """

    # make input/output files absolute
    args.infile  = [op.realpath(f) for f in args.infile]
    args.outfile =  op.realpath(args.outfile)

    # the importing.loadData function accepts
    # either a single encoding, or one encoding
    # for each data file. If the former, we
    # convert it into the latter, and then we
    # convert it into a dict of {infile:encoding}
    # mappings
    if args.encoding is not None:
        if len(args.encoding) == 1:
            args.encoding = [args.encoding[0]] * len(args.infile)
        elif len(args.encoding) != len(args.infile):
            raise ValueError('Wrong number of encodings specified - specify '
                             'either one encoding, or one encoding for each '
                             'input file.')
        args.encoding = dict(zip(args.infile, args.encoding))

    # turn loaders into dict of { absfile : name } mappings
    if args.loader is not None:
        args.loader = {op.realpath(f) : n for f, n in args.loader}

    # Remove/rename duplicates options are mutually exclusive
    if args.remove_duplicates and args.rename_duplicates:
        raise ValueError('Only one of--remove_duplicates and '
                         '--rename_duplicates may be used.')

    # turn index indices into dict of
    # { file : [index] } mappings
    if args.index is not None:
        indexes = {}
        for fname, idx in args.index:
            idx = [int(i) for i in idx.split(',')]
            indexes[op.abspath(fname)] = idx
        args.index = indexes

    # turn formatters into dict of { vid : name } mappings
    if args.var_format is None:
        args.var_format = []

    var_format = {}
    for v, n in args.var_format:

        # Formatters should be set on integer
        # variable IDs. But we allow non-integers
        # to pass through, as the exportData
        # function will also check against column
        # names.
        try:               v = int(v)
        except ValueError: pass

        var_format[v] = n

    args.var_format = var_format


def _prepareAuxFileNames(args):
    """Sub-function of :func:`parseArgs`. Prepares arguments related to
    the auxillary output files.

    :arg args: ``argparse.Namespace`` object.
    """
    # assign default names to
    # auxillary output files
    outbase, outext = op.splitext(args.outfile)
    auxfiles        = ['log', 'non_numerics', 'unknown_vars',
                       'icd10_map', 'description', 'summary']
    auxexts         = collections.defaultdict(lambda : '.txt',
                                              non_numerics=outext)
    for auxfile in auxfiles:
        writeatt = 'write_{}'.format(auxfile)
        fileatt  = '{}_file' .format(auxfile)
        write    = getattr(args, writeatt)
        filename = getattr(args, fileatt)

        if write and filename is None:
            setattr(args, fileatt, '{}_{}{}'.format(outbase,
                                                    auxfile,
                                                    auxexts[auxfile]))
        else:
            setattr(args, writeatt, filename is not None)


def _prepareSubjectAndVariableSelectors(args):
    """Sub-function of :func:`parseArgs`. Prepares arguments related to
    the subject and variable selection options.

    :arg args: ``argparse.Namespace`` object.
    """

    # turn --subject/--variable/--exclude/
    # --exclude_variable arguments into
    # lists of IDs. If error is True, an
    # error is raised on unparseable
    # arguments.
    def replaceIDs(things, error=True):
        newthings = []
        failed    = []
        for thing in things:

            # --subject/--variable/--exclude args
            # may be may be a file name containing
            # a list of IDs
            if op.exists(thing):
                with open(thing, 'rt') as f:
                    parsed = f.read().split()
                    parsed = [int(t.strip()) for t in parsed]

            else:
                # Or they may be one or more comma-separated
                # IDs or matlab start:step[:stop] ranges,
                # both handled by the parseMatlabRange function.
                try:
                    parsed = []
                    for tkn in thing.split(','):
                        parsed.extend(util.parseMatlabRange(tkn))

                except ValueError:
                    parsed = None

                    # --subject may also be an expression,
                    # so if error is False, and the range/
                    # list parses fail, we pass the argument
                    # through. Otherwise we propagate the
                    # error.
                    if error:
                        raise

                if parsed is None:
                    failed.append(thing)
                    continue

            for t in parsed:
                if t not in newthings:
                    newthings.append(t)

        if len(newthings) == 0: newthings = None
        if len(failed)    == 0: failed    = None

        return newthings, failed

    # variable/exclude is transformed into
    # a list of integer IDs, but subject is
    # transformed into a tuple containing
    # ([ID], [exprStr])
    if args.subject is not None:
        args.subject  = replaceIDs(args.subject, False)
    else:
        args.subject  = None, None

    if args.variable is not None:
        args.variable = replaceIDs(args.variable)[0]
    if args.exclude is not None:
        args.exclude = replaceIDs(args.exclude)[0]
    if args.exclude_variable is not None:
        args.exclude_variable  = replaceIDs(args.exclude_variable)[0]


def _prepareCategorySelectors(args):
    """Sub-function of :func:`parseArgs`. Prepares arguments related to
    the category selection options.

    :arg args: ``argparse.Namespace`` object.
    """

    # categories can be specified
    # either by name or by ID -
    # convert the latter to integers.
    if args.category is not None:
        for i, c in enumerate(args.category):
            try:               args.category[i] = int(c)
            except ValueError: continue
    if args.exclude_category is not None:
        for i, c in enumerate(args.exclude_category):
            try:               args.exclude_category[i] = int(c)
            except ValueError: continue


def _prepareColumnSelectors(args):
    """Sub-function of :func:`parseArgs`. Prepares arguments related to
    the column selection options.

    :arg args: ``argparse.Namespace`` object.
    """
    # The column option accepts
    # column names, or a file
    # containing column names
    def loadIfExists(path):
        if op.exists(path):
            with open(path, 'rt') as f:
                items = f.readlines()
        else:
            items = [path]
        return [i.strip() for i in items]

    if args.column is not None:
        args.column = list(it.chain(*[loadIfExists(c) for c in args.column]))


def _prepareVisitSelectors(args):
    """Sub-function of :func:`parseArgs`. Prepares arguments related to
    visit selection options.

    :arg args: ``argparse.Namespace`` object.
    """
    # visits are restricted using the
    # keepVisits cleaning function
    if args.visit is not None:
        visit = []
        for v in args.visit:
            if v in ('first', 'last'): visit.append('"{}"'.format(v))
            else:                      visit.append(str(v))

        visit = 'keepVisits({})'.format(','.join(visit))

        if args.global_clean is None: args.global_clean  =       visit
        else:                         args.global_clean += ',' + visit


def _prepareCleaningSelectors(args):
    """Sub-function of :func:`parseArgs`. Prepares arguments related to
    cleaning function options.

    :arg args: ``argparse.Namespace`` object.
    """

    # convert na_values from a sequence of [(varid, values)]
    # pairs into a dict of {varid : values} mappings
    if args.na_values is not None:
        args.na_values = {int(vid) : values for vid, values in args.na_values}

    # Convert recoding from a sequence of [(varid, rawlevels, newlevels)]
    # tuples to a dict of {varid : (rawlevels, newlevels)} mappings
    if args.recoding is not None:
        args.recoding = {int(vid) : (rawlevels, newlevels)
                         for vid, rawlevels, newlevels in args.recoding}

    # Convert child_values from a sequence of
    # [(varid, exprs, values)] tuples to a dict of
    # {varid : (exprs, values)} mappings
    if args.child_values is not None:
        args.child_values = {int(vid) : (exprs, values)
                             for vid, exprs, values in args.child_values}

    # convert clean from a sequence of [(varid, expr)]
    # pairs into a dict of {varid : expr} mappings.
    if args.clean is not None:
        args.clean = {int(vid) : expr for vid, expr in args.clean}

    # convert type_clean from a sequence of [(type, expr)]
    # pairs into a dict of {type : expr} mappings.
    if args.type_clean is not None:
        args.type_clean = {util.CTYPES[t] : e for t, e in args.type_clean}


def _prepareIDsOnly(args):
    """Sub-function of :func:`parseArgs`. Prepares arguments related to
    the --ids_only option.

    :arg args: ``argparse.Namespace`` object.
    """
    if not args.ids_only:
        return

    args.skip_insertna      = True
    args.skip_childvalues   = True
    args.skip_clean_funcs   = True
    args.skip_recoding      = True
    args.skip_processing    = True

    # See main.main, and the do* functions
    # that it calls - aux file export is
    # ultimately toggled by these arguments
    args.unknown_vars_file  = None
    args.icd10_map_file     = None
    args.description_file   = None
    args.summary_file       = None



#!/usr/bin/env python
#
# test_importing_removeSubjects.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

import textwrap as tw
import pytest

import pandas as pd
import numpy as np

import funpack.importing.filter as filter

from funpack.datatable import Column
import funpack.fileinfo       as fileinfo
import funpack.processing     as processing
from . import (gen_DataTable, gen_DataTableFromDataFrame,
               tempdir, gen_tables, gen_test_data, clear_plugins)


def test_columnsToLoad():

    data = tw.dedent("""
    eid,1-0.0,2-0.0,3-0.0
    1,10,20,30
    2,11,21,31
    3,12,22,32
    4,13,23,33
    """).strip()

    vartable = gen_tables(range(1, 10))[0]

    with tempdir():

        with open('data.txt', 'wt') as f:
            f.write(data)

        finfo = fileinfo.FileInfo('data.txt')
        cols  = finfo.columns('data.txt')

        gc, gd = filter.columnsToLoad(finfo, vartable, None)
        assert gc['data.txt'] == cols
        assert gd             == []

        gc, gd = filter.columnsToLoad(finfo, vartable, [1, 2, 3])
        assert gc['data.txt'] == cols[:4]
        assert gd             == []

        gc, gd = filter.columnsToLoad(finfo, vartable, None, exclude=[2, 3])
        assert gc['data.txt'] == cols[:2]
        assert gd             == cols[2:]

        gc, gd = filter.columnsToLoad(finfo, vartable, [1, 2, 3], exclude=[2, 3])
        assert gc['data.txt'] == cols[:2]
        assert gd             == cols[2:]


def test_filterSubjects():

    data       = np.random.randint(1, 10, (500, 11))
    index      = np.arange(1, 501)
    data[:, 0] = index
    colnames   = ['eid'] + ['{}-0.0'.format(i+1) for i in range(10)]
    cols       = [Column(None, n, i, i) for i, n in enumerate(colnames)]
    df         = pd.DataFrame(data, columns=colnames).set_index('eid')
    data       = data[:, 1:]

    #   - exclude:  list of subjects to *exclude*
    #   - include:  list of subjects to *include*
    #   - exprs:    list of expressions specifying *inclusion*

    # pass through
    got = filter.filterSubjects(df, cols)
    assert np.all(got.index == np.arange(1, 501))
    assert np.all(got == df)

    # include
    got = filter.filterSubjects(df, cols, subjects=np.arange(1, 20))
    assert np.all(got.index == np.arange(1, 20))
    assert np.all(got == df.loc[np.arange(1, 20), :])

    # exclude
    got = filter.filterSubjects(df, cols, exclude=[1, 2, 3])
    assert np.all(got.index == np.arange(4, 501))
    assert np.all(got == df.loc[np.arange(4, 501), :])

    # expr
    mask = data[:, 0] > 5
    got = filter.filterSubjects(df, cols, subjectExprs=['v1 > 5'])
    assert np.all(got.index == (np.where(mask)[0] + 1))
    assert np.all(got == data[mask, :])

    # include + exclude
    got = filter.filterSubjects(
        df, cols, subjects=np.arange(1, 20), exclude=[1, 2, 3])
    assert np.all(got.index == np.arange(4, 20))
    assert np.all(got == df.loc[np.arange(4, 20), :])

    # include + expr
    mask       = data[:, 0] > 5
    mask[100:] = 0
    got = filter.filterSubjects(
        df, cols, subjects=np.arange(1, 101), subjectExprs=['v1 > 5'])
    assert np.all(got.index == (np.where(mask)[0] + 1))
    assert np.all(got == data[mask, :])

    # expr + exclude
    mask     = data[:, 0] > 5
    mask[:3] = 0
    got = filter.filterSubjects(
        df, cols, subjectExprs=['v1 > 5'], exclude=[1, 2, 3])
    assert np.all(got.index == (np.where(mask)[0] + 1))
    assert np.all(got == data[mask, :])

    # include + expr + exclude
    mask          = data[:, 0] > 5
    mask[300:]    = 0
    mask[75:200]  = 0
    got = filter.filterSubjects(
        df, cols,
        subjects=np.arange(1, 301),
        exclude=np.arange(76, 201),
        subjectExprs=['v1 > 5'])
    assert np.all(got.index == (np.where(mask)[0] + 1))
    assert np.all(got == data[mask, :])

    # multiple expressions
    got = filter.filterSubjects(
        df, cols, subjectExprs=['v1 > 5', 'v2 == 9'])
    mask = (data[:, 0] > 5) | (data[:, 1] == 9)
    assert np.all(got.index == np.where(mask)[0] + 1)
    assert np.all(got == data[mask, :])


def test_filterSubjects_multiple_columns():

    def gendata():
        colnames   = ['eid', '1-0.0', '1-1.0', '1-2.0', '2-0.0', '2-1.0']
        variables  = [0,      1,       1,       1,       2,       2]
        cols       = [Column(None, n, i, v) for i, (n, v) in
                      enumerate(zip(colnames, variables))]
        data       = np.random.randint(1, 10, (6, 500))
        data[0, :] = np.arange(1, 501)
        df         = pd.DataFrame({c : d for c, d in zip(colnames, data)})
        df         = df.set_index('eid')
        data       = data[1:, :].T
        return df, cols, data

    def all(s): return s.all(axis=1)
    def any(s): return s.any(axis=1)

    # combine vars with ncolumn
    # mismatch - should be ORed
    # within var, then combined
    df, cols, data = gendata()
    exprs = ['v1 > 2 && v2 < 7']
    exp   = any(data[:, :3] > 2) & any(data[:, 3:] < 7)
    got   = filter.filterSubjects(df, cols, subjectExprs=exprs)
    assert (got.index == (np.where(exp)[0] + 1)).all()

    # combine columns within var
    df, cols, data = gendata()
    exprs = ['all(v1 > 2) && any(v2 < 7)']
    exp = all(data[:, :3] > 2) & any(data[:, 3:] < 7)
    got = filter.filterSubjects(df, cols, subjectExprs=exprs)
    assert (got.index == (np.where(exp)[0] + 1)).all()

    # no combining columns - should
    # default to any
    df, cols, data = gendata()
    exprs = ['v1 > 6']
    exp = any(data[:, :3] > 6)
    got = filter.filterSubjects(df, cols, subjectExprs=exprs)
    assert (got.index == (np.where(exp)[0] + 1)).all()

    # multipler expressions - ORed together
    df, cols, data = gendata()
    exprs = ['v1 > 6', 'v2 < 4']
    exp = any(data[:, :3] > 6) | any(data[:, 3:] < 4)
    got = filter.filterSubjects(df, cols, subjectExprs=exprs)
    assert (got.index == (np.where(exp)[0] + 1)).all()


def test_filter_subjects_column_not_present():
    data       = np.random.randint(1, 10, (500, 11))
    index      = np.arange(1, 501)
    data[:, 0] = index
    colnames   = ['eid'] + ['{}-0.0'.format(i+1) for i in range(10)]
    cols       = [Column(None, n, i, i) for i, n in enumerate(colnames)]
    df         = pd.DataFrame(data, columns=colnames).set_index('eid')

    # expression is skipped if any variable is not present
    got1       = filter.filterSubjects(df, cols, subjectExprs=['v11 > 0'])
    got2       = filter.filterSubjects(df, cols, subjectExprs=['v9  < 0 || '
                                                               'v11 > 0'])

    assert (got1 == df).all().all()
    assert (got2 == df).all().all()


@clear_plugins
def test_addAuxillaryVariables():
    data = tw.dedent("""
    eid,1-0.0,2-0.0,3-0.0
    1,10,20,30
    2,11,21,31
    3,12,22,32
    4,13,23,33
    """).strip()

    proctable = pd.DataFrame({
        'ID'       : [0],
        'Variable' : [['vids', [1]]],
        'Process'  : [{'binariseCategorical' :
                      processing.Process('processor',
                                         'binariseCategorical',
                                         (),
                                         dict(take=[2]),
                                         '')}]})

    with tempdir():

        with open('data.txt', 'wt') as f:
            f.write(data)
        finfo = fileinfo.FileInfo('data.txt')

        assert filter.addAuxillaryVariables(
            finfo, proctable, None, None) == (None, None)
        assert filter.addAuxillaryVariables(
            finfo, proctable, [1, 2, 3], None) == ([1, 2, 3], None)
        assert filter.addAuxillaryVariables(
            finfo, proctable, [1, 3], None) == ([1, 3, 2], None)
        assert filter.addAuxillaryVariables(
            finfo, proctable, [1, 2, 3], [2]) == ([1, 2, 3], None)
        assert filter.addAuxillaryVariables(
            finfo, proctable, [1, 3], [2]) == ([1, 3, 2], None)
        assert filter.addAuxillaryVariables(
            finfo, proctable, [1, 3], [2, 3]) == ([1, 3, 2], [3])



#!/usr/bin/env python
#
# test_importing_mergeData.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

import itertools as it
import textwrap  as tw

import pandas as pd
import numpy  as np

import funpack.importing as importing
import funpack.merging   as merging
import funpack.fileinfo  as fileinfo

from . import tempdir, patch_logging


def _merge_testdata(samesubjs=False, samecols=False):

    if not samecols:
        headers = [
            'eid 1-0.0 2-0.0 3-0.0',
            'eid 1-0.0 2-0.0 3-0.0 4-0.0',
            'subj one two three',
            None]
    else:
        headers = [
            'eid 1-0.0 2-0.0 3-0.0',
            'eid 2-0.0 3-0.0 4-0.0',
            'eid 1-0.0 3-0.0 4-0.0',
            'eid 2-0.0 3-0.0 4-0.0 5-0.0']

    files = []
    data  = []

    for i, hdr in enumerate(headers):
        if hdr is not None:
            # we make one file arbitrarily
            # bigger than the rest
            ncols = len(hdr.split())
            if ncols == 5:
                nrows = 3
            else:
                nrows = 2
        else:
            ncols = 4
            nrows = 2

        fdata = []
        for ri in range(nrows):

            if samesubjs: row = [1000 + ri]
            else:         row = [(i + 1) * 1000 + ri]
            for ci in range(ncols - 1):
                row.append((i + 1) * 100 + ri * 10 + ci)
            fdata.append(row)
        files.append('test{}.txt'.format(i))
        with open(files[-1], 'wt') as f:
            if hdr is not None:
                f.write(hdr + '\n')
            for r in fdata:
                f.write(' '.join([str(c) for c in r]) + '\n')

    cols = fileinfo.fileinfo(files)[2]
    for fcols, hdr, f in zip(cols, headers, files):
        data.append(pd.read_csv(f,
                                index_col=0,
                                names=[c.name for c in fcols],
                                header=None if hdr is None else 0,
                                delim_whitespace=True))

    return files, data, cols


@patch_logging
def test_mergeData_cols_naive():

    with tempdir():
        files, data, cols = _merge_testdata()

    datacopy = [d.copy() for d in data]

    merged, mcols = merging.mergeDataFrames(datacopy, cols, 'cols', 'naive')

    # Row indices are taken from
    # the file with the most rows
    lens   = [len(d.index) for d in data]
    maxlen = max(lens)
    lidx   = lens.index(maxlen)

    expcols = [cols[lidx][0]] + list(it.chain(*[c[1:] for c in cols]))
    exprows = data[lidx].index.values

    assert mcols == expcols
    assert np.all(merged.index.values == exprows)
    assert np.all(merged.columns == [c.name for c in mcols[1:]])

    for i, col in enumerate(mcols[1:]):

        didx    = files.index(col.datafile)
        cmdata  = merged.iloc[:, i].values

        expdata        = np.empty(maxlen, dtype=np.float)
        clen           = len(data[didx].iloc[:, col.index - 1])
        expdata[:clen] =     data[didx].iloc[:, col.index - 1].values
        expdata[clen:] = np.nan

        cmnan  = np.isnan(cmdata)
        expnan = np.isnan(expdata)

        assert np.all(cmdata[~cmnan] == expdata[~expnan])
        assert np.all(cmnan          == expnan)


@patch_logging
def test_mergeData_rows_naive():

    with tempdir():
        files, data, cols = _merge_testdata()

    datacopy = [d.copy() for d in data]

    merged, mcols = merging.mergeDataFrames(datacopy, cols, 'rows', 'naive')

    # columns are taken from the
    # file with the most columns
    lens   = [len(d.columns) for d in data]
    maxlen = max(lens)
    lidx   = lens.index(maxlen)

    expcols = cols[lidx]
    exprows = list(it.chain(*[d.index.values for d in data]))
    nrows   = len(exprows)


    assert mcols == expcols
    assert np.all(merged.index.values == exprows)
    assert np.all(merged.columns == [c.name for c in mcols[1:]])


    for i, col in enumerate(mcols[1:]):

        cmdata     = merged.iloc[:, i].values
        expdata    = np.empty(nrows, dtype=np.float)
        expdata[:] = np.nan

        rowoff = 0
        for d in data:
            if (col.index - 1) < len(d.columns):
                dcd = d.iloc[:, col.index - 1]
                expdata[rowoff:rowoff + len(dcd)] = dcd
            rowoff += len(d.index)

        cmnan  = np.isnan(cmdata)
        expnan = np.isnan(expdata)

        assert np.all(cmdata[~cmnan] == expdata[~expnan])
        assert np.all(cmnan          == expnan)



@patch_logging
def test_mergeData_cols_inner():

    with tempdir():
        files, data, cols = _merge_testdata(samesubjs=True)

    datacopy = [d.copy() for d in data]

    merged, mcols = merging.mergeDataFrames(datacopy, cols, 'cols', 'inner')

    # only rows for subjects which
    # are present in all files should
    # be present in the merged data.
    expcols = [cols[0][0]] + list(it.chain(*[c[1:] for c in cols]))
    exprows = list(sorted(set.intersection(*[set(d.index.values)
                                             for d in data])))

    assert mcols == expcols
    assert np.all(merged.index.values == exprows)
    assert np.all(merged.columns == [c.name for c in mcols[1:]])

    for rid in exprows:
        row = []

        for d in data:
            if rid not in d.index:
                continue

            row.extend(list(d.loc[rid, :].values))

        assert np.all(merged.loc[rid, :] == row)


@patch_logging
def test_mergeData_cols_outer():

    with tempdir():
        files, data, cols = _merge_testdata(samesubjs=True)

    datacopy = [d.copy() for d in data]

    merged, mcols = merging.mergeDataFrames(datacopy, cols, 'cols', 'outer')

    # rows for subjects which are
    # present in any files should
    # be present in the merged data.
    expcols = [cols[0][0]] + list(it.chain(*[c[1:] for c in cols]))
    exprows = list(sorted(set.union(*[set(d.index.values)
                                      for d in data])))

    assert mcols == expcols
    assert np.all(merged.index.values == exprows)
    assert np.all(merged.columns == [c.name for c in mcols[1:]])

    for rid in exprows:
        row = []

        for d in data:
            if rid not in d.index:
                row.extend([np.nan] * len(d.columns))
                continue

            row.extend(list(d.loc[rid, :].values))

        exp    = np.array(row)
        got    = merged.loc[rid, :]
        expnan = np.isnan(exp)
        gotnan = np.isnan(got)

        assert np.all(gotnan       == expnan)
        assert np.all(got[~gotnan] == exp[~expnan])


@patch_logging
def test_mergeData_rows_inner():

    with tempdir():
        files, data, cols = _merge_testdata(samecols=True)

    datacopy = [d.copy() for d in data]

    merged, mcols = merging.mergeDataFrames(datacopy, cols, 'rows', 'inner')

    # expected columns are those
    # which are present in all
    # files
    colnames = [[c.name for c in cc[1:]] for cc in cols]
    joined   = set.intersection(*[set(c) for c in colnames])
    allcols  = list(it.chain(*[c[1:] for c in cols]))
    colnames = list(it.chain(*colnames))

    colidxs = [colnames.index(j) for j in joined]
    expcols = [cols[0][0]] + [allcols[i] for i in colidxs]

    assert np.all(merged.columns == [ec.name for ec in expcols[1:]])
    assert mcols == expcols

    for col in mcols[1:]:
        coldata = merged.loc[:, col.name].values
        expdata = []
        for d in datacopy:
            expdata.extend(list(d.loc[:, col.name].values))
        assert np.all(coldata == expdata)


@patch_logging
def test_mergeData_rows_outer():

    with tempdir():
        files, data, cols = _merge_testdata(samecols=True)

    datacopy = [d.copy() for d in data]

    merged, mcols = merging.mergeDataFrames(datacopy, cols, 'rows', 'outer')

    # all columns should be
    # present, in the correct
    # order
    uniqnames = []
    uniqcols  = []

    for col in it.chain(*[c[1:] for c in cols]):
        if col.name not in uniqnames:
            uniqnames.append(col.name)
            uniqcols .append(col)
    expcols = [cols[0][0]] + uniqcols

    exprows = list(it.chain(*[d.index for d in datacopy]))

    assert np.all(merged.index.values == exprows)
    assert np.all(merged.columns == [ec.name for ec in expcols[1:]])
    assert mcols == expcols

    for col in mcols[1:]:
        got = merged.loc[:, col.name].values
        exp = []

        for d in datacopy:
            if col.name in d.columns:
                exp += list(d.loc[:, col.name].values)

            else:
                exp += [np.nan] * len(d.index)

        exp    = np.array(exp)
        gotnan = np.isnan(got)
        expnan = np.isnan(exp)

        assert np.all(gotnan == expnan)
        assert np.all(got[~gotnan] == exp[~expnan])



@patch_logging
def test_mergeData_non0_indexes_cols():

    data1 = tw.dedent("""
    eid,col1,col2
    1,11,21
    2,12,22
    3,13,23
    """)
    data2 = tw.dedent("""
    col3,eid,col4
    31,1,41
    32,2,42
    33,3,43
    """)
    data3 = tw.dedent("""
    col5,col6,eid
    51,61,1
    52,62,2
    53,63,3
    """)

    with tempdir():
        with open('data1.txt', 'wt') as f: f.write(data1)
        with open('data2.txt', 'wt') as f: f.write(data2)
        with open('data3.txt', 'wt') as f: f.write(data3)

        cols = fileinfo.fileinfo(['data1.txt', 'data2.txt', 'data3.txt'],
                                 indexes={'data1.txt' : [0],
                                          'data2.txt' : [1],
                                          'data3.txt' : [2]})[2]

        df1 = pd.read_csv('data1.txt', delimiter=',', index_col=0)
        df2 = pd.read_csv('data2.txt', delimiter=',', index_col=1)
        df3 = pd.read_csv('data3.txt', delimiter=',', index_col=2)

        merged, gotcols = merging.mergeDataFrames([df1, df2, df3],
                                                  cols,
                                                  axis='variables',
                                                  strategy='inner')


        expcols = cols[0] + [cols[1][0], cols[1][2], cols[2][0], cols[2][1]]

        assert gotcols == expcols
        assert np.all(merged.columns == [e.name for e in expcols[1:]])
        assert np.all(merged.index   == [1,  2,  3])
        assert np.all(merged['col1'] == [11, 12, 13])
        assert np.all(merged['col2'] == [21, 22, 23])
        assert np.all(merged['col3'] == [31, 32, 33])
        assert np.all(merged['col4'] == [41, 42, 43])
        assert np.all(merged['col5'] == [51, 52, 53])
        assert np.all(merged['col6'] == [61, 62, 63])


@patch_logging
def test_mergeData_non0_indexes_rows():

    data1 = tw.dedent("""
    eid,col1,col2
    1,11,21
    2,12,22
    3,13,23
    """)
    data2 = tw.dedent("""
    col1,eid,col2
    14,4,24
    15,5,25
    16,6,26
    """)
    data3 = tw.dedent("""
    col1,col2,eid
    17,27,7
    18,28,8
    19,29,9
    """)

    with tempdir():

        with open('data1.txt', 'wt') as f: f.write(data1)
        with open('data2.txt', 'wt') as f: f.write(data2)
        with open('data3.txt', 'wt') as f: f.write(data3)

        cols = fileinfo.fileinfo(['data1.txt', 'data2.txt', 'data3.txt'],
                                 indexes={'data1.txt' : [0],
                                          'data2.txt' : [1],
                                          'data3.txt' : [2]})[2]

        df1 = pd.read_csv('data1.txt', delimiter=',', index_col=0)
        df2 = pd.read_csv('data2.txt', delimiter=',', index_col=1)
        df3 = pd.read_csv('data3.txt', delimiter=',', index_col=2)

        merged, gotcols = merging.mergeDataFrames([df1, df2, df3],
                                                  cols,
                                                  axis='subjects',
                                                  strategy='inner')

        expcols = cols[0]

        assert gotcols == expcols
        assert np.all(merged.columns == [e.name for e in expcols[1:]])
        assert np.all(merged.index   == np.arange( 1, 10))
        assert np.all(merged['col1'] == np.arange(11, 20))
        assert np.all(merged['col2'] == np.arange(21, 30))


@patch_logging
def test_mergeData_different_index_names():
    data1 = tw.dedent("""
    id1,col1,col2
    1,14,24
    2,15,25
    3,16,26
    """)
    data2 = tw.dedent("""
    id2,col1,col2
    1,17,27
    2,18,28
    3,19,29
    """)

    with tempdir():
        with open('data1.txt', 'wt') as f: f.write(data1)
        with open('data2.txt', 'wt') as f: f.write(data2)

        cols = fileinfo.fileinfo(['data1.txt', 'data2.txt'])[2]

        df1 = pd.read_csv('data1.txt', delimiter=',')
        df2 = pd.read_csv('data2.txt', delimiter=',')

        merged, gotcols = merging.mergeDataFrames([df1, df2],
                                                  cols,
                                                  axis='subjects',
                                                  strategy='inner')

        assert gotcols[0] == cols[0][0]
        assert merged.index.name == 'id1'



#!/usr/bin/env python
#
# test_nonstandard_columns.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

import             shlex
import textwrap as tw

import pandas   as pd
import numpy    as np

import funpack.main as main


from . import (patch_logging,
               clear_plugins,
               tempdir,
               gen_tables,
               gen_DataTable,
               gen_test_data,
               gen_DataTableFromDataFrame)


@patch_logging
def test_nonstandard_columns():

    indata = tw.dedent("""
    f.eid,f.1.0.0,f.2.0.0,f.3.0.0
    1,4,7,3
    2,1,9,2
    3,,3,5
    """).strip()

    tests = [
        ('-nb -nv 2 7',
         """
         f.eid,f.1.0.0,f.2.0.0,f.3.0.0
         1,4,,3
         2,1,9,2
         3,,3,5
         """),
        ('-nb -re 3 2,3 10,20',
         """
         f.eid,f.1.0.0,f.2.0.0,f.3.0.0
         1,4,7,20
         2,1,9,10
         3,,3,5
         """),
        ('-nb -cv 1 "v2 == 3" 25',
         """
         f.eid,f.1.0.0,f.2.0.0,f.3.0.0
         1,4,7,3
         2,1,9,2
         3,25,3,5
         """),
    ]

    for cmd, expout in tests:
        with tempdir():
            expout = tw.dedent(expout).strip()

            with open('in.csv',  'wt') as f: f.write(indata)
            with open('exp.csv', 'wt') as f: f.write(expout)

            main.main(shlex.split(cmd + ' out.tsv in.csv'))

            got = pd.read_csv('out.tsv', index_col=0, delimiter='\t')
            exp = pd.read_csv('exp.csv', index_col=0)

            assert np.all(exp.columns             == got.columns)
            assert np.all(np.isnan(exp.loc[:, :]) == np.isnan(got.loc[:, :]))
            assert np.all(exp.loc[:, :].notna()   == got.loc[:, :].notna())



#!/usr/bin/env python
#
# test_exporting.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#


import             warnings
import             os
import textwrap as tw
import os.path  as op
import multiprocessing as mp
from unittest import mock

import pytest

from collections import OrderedDict

import pandas as pd
import numpy  as np
import           h5py

import funpack.importing      as importing
import funpack.exporting      as exporting
import funpack.exporting_hdf5 as exporting_hdf5
import funpack.exporting_tsv  as exporting_tsv
import funpack.datatable      as datatable
import funpack.fileinfo       as fileinfo
import funpack.custom         as custom
import funpack.util           as util

from . import (gen_DataTable,
               clear_plugins,
               patch_base_tables,
               tempdir,
               gen_test_data,
               gen_tables,
               gen_DataTableFromDataFrame)


def test_exportData():

    custom.registerBuiltIns()

    testdata = np.random.randint(1, 10, (10, 10)).astype(np.float32)

    def check(gotfile, expected):
        with open(gotfile, 'rt') as f:
            got = f.read().strip()
            return got == expected

    with tempdir():

        dtable = gen_DataTable(testdata)
        exporting.exportData(dtable, 'data.tsv', 'tsv')
        exp = ['\t'.join(['eid'] +
                         ['{}-0.0'.format(i) for i in range(1, 11)])] + \
              ['\t'.join([str(i + 1)] + [str(c) for c in r])
               for i, r in enumerate(testdata.T)]
        exp = '\n'.join(exp)
        assert check('data.tsv', exp)

        td = np.copy(testdata)
        td[5, 5] = np.nan
        dtable = gen_DataTable(td)
        exporting.exportData(dtable,
                             'data.tsv',
                             'tsv',
                             sep='*',
                             missingValues='boo')
        exp = ['*'.join(['eid'] +
                         ['{}-0.0'.format(i)  for i in range(1, 11)])] + \
              ['*'.join([str(i + 1)] +
                        ['boo' if np.isnan(c) else str(c) for c in r])
               for i, r in enumerate(td.T)]
        exp = '\n'.join(exp)
        assert check('data.tsv', exp)


def test_exportData_subjid():

    custom.registerBuiltIns()

    with tempdir():
        gen_test_data(3, 3, 'data.tsv', start_subj=10)

        variables = list(range(0, 4))
        vartable, proctable, cattable, _ = gen_tables(variables)
        colnames  = ['eid'] + ['{}-0.0'.format(v) for v in variables[1:]]
        colobjs   = [datatable.Column(None, c, i, v, 0, 0)
                     for i, (c, v) in enumerate(zip(colnames, variables))]
        data      = pd.read_csv( 'data.tsv', delimiter='\t', index_col=0)

        dtable = datatable.DataTable(
            data, colobjs, vartable, proctable, cattable)

        exporting.exportData(dtable, 'export.tsv', 'tsv')

        got = pd.read_csv('export.tsv', delimiter='\t', index_col=0)

        assert np.all(got.index == [10, 11, 12])



def test_exportData_TSV_subject_ordering():

    custom.registerBuiltIns()

    with tempdir():
        data   = np.random.randint(1, 10, (5, 20))
        dtable = gen_DataTable(data)
        subjs  = np.arange(20, 0, -1)
        dtable = dtable.subtable(rows=subjs)

        exporting.exportData(dtable, 'data.tsv', 'tsv')
        got = pd.read_csv('data.tsv', delimiter='\t', index_col=0)
        assert np.all(got.index == subjs)

        exporting.exportData(dtable,
                             'data.tsv',
                             'tsv',
                             numRows=2)
        got = pd.read_csv('data.tsv', delimiter='\t', index_col=0)
        assert np.all(got.index == subjs)


def test_exportData_numRows():

    custom.registerBuiltIns()

    with tempdir():
        data   = np.random.randint(1, 10, (5, 20))
        dtable = gen_DataTable(data)

        exporting.exportData(dtable, 'nochunks.tsv', 'tsv')
        exporting.exportData(dtable, 'chunks.tsv', 'tsv', numRows=2)

        with open('nochunks.tsv', 'rt') as f: nochunks = f.read()
        with open('chunks.tsv',   'rt') as f: chunks   = f.read()

        assert nochunks == chunks


def test_exportData_dropNaRows():

    custom.registerBuiltIns()

    with tempdir():
        data   = np.random.random((10, 10))
        data[:, [2, 4, 7]] = np.nan
        dtable = gen_DataTable(data)

        exporting.exportData(dtable, 'data.tsv', 'tsv',  dropNaRows=True)
        exporting.exportData(dtable, 'data.h5',  'hdf5', dropNaRows=True)

        got1 = pd.read_csv('data.tsv', sep='\t', index_col=0)
        got2 = pd.read_hdf('data.h5')
        exp  = data[:, [0, 1, 3, 5, 6, 8, 9]].T

        for got in [got1, got2]:
            assert list(got.index) == [1, 2, 4, 6, 7, 9, 10]
            assert np.all(np.isclose(exp, got.values))


def test_defaultDateTimeFormat():

    custom.registerBuiltIns()

    with tempdir():
        gen_test_data(3, 3, 'data.tsv', ctypes={1 : 'date', 2 : 'datetime'})

        data = pd.read_csv('data.tsv',
                           delimiter='\t',
                           parse_dates=['1-0.0', '2-0.0'],
                           infer_datetime_format=True,
                           index_col=0)
        dt = gen_DataTableFromDataFrame(data)

        dt.vartable.loc[1, 'Type'] = util.CTYPES.date
        dt.vartable.loc[2, 'Type'] = util.CTYPES.time

        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            datecol        = dt[:, '1-0.0']
            datecol[2]     = np.nan
            dt[:, '1-0.0'] = datecol

        def datefmt(d):
            try:    return d.strftime('%Y-%m-%d')
            except: return np.nan
        def timefmt(t):
            try:    return t.strftime('%Y-%m-%dT%H:%M:%S%z')
            except: return np.nan

        expdate = data['1-0.0'].apply(datefmt)
        exptime = data['2-0.0'].apply(timefmt)
        expvar  = data['3-0.0']

        gotdate = exporting.defaultDateTimeFormat(
            dt, dt.columns(1)[0], data['1-0.0'])
        gottime = exporting.defaultDateTimeFormat(
            dt, dt.columns(2)[0], data['2-0.0'])
        gotvar  = exporting.defaultDateTimeFormat(
            dt, dt.columns(3)[0], data['3-0.0'])

        expdna = expdate.isna()
        exptna = exptime.isna()

        assert np.all(expdate.isna() == expdna)
        assert np.all(exptime.isna() == exptna)

        assert np.all(expdate[~expdna] == gotdate[~expdna])
        assert np.all(exptime[~exptna] == gottime[~exptna])
        assert np.all(expvar           == gotvar)


def test_exportHDF5():

    custom.registerBuiltIns()

    with tempdir():
        gen_test_data(5, 10, 'data.tsv', ctypes={1 : 'date', 2 : 'datetime'})

        data = pd.read_csv('data.tsv',
                           delimiter='\t',
                           parse_dates=['1-0.0', '2-0.0'],
                           infer_datetime_format=True,
                           index_col=0)
        dt = gen_DataTableFromDataFrame(data)

        for col in dt.dataColumns:
            exporting.formatColumn(col,
                                   dt,
                                   'default',
                                   'default',
                                   {})

        exporting_hdf5.exportHDF5(dt,
                                  'out_funpack.h5',
                                  key='h5key',
                                  style='funpack')
        exporting_hdf5.exportHDF5(dt,
                                  'out_pandas.h5',
                                  key='h5key',
                                  style='pandas')

        colnames = [c.name for c in dt.dataColumns]

        exp    = dt[:, :]
        gotpd  = pd.read_hdf('out_pandas.h5')
        gotukb = h5py.File('out_funpack.h5', 'r')
        assert gotpd.index.name     == 'eid'
        assert np.all(gotpd.columns == colnames)

        for i, c in enumerate(exp.columns):
            assert np.all(gotpd[colnames[i]] == exp[c])

        gotidx = gotukb['h5key/eid']

        assert np.all(exp.index[:] == gotidx)

        h5ver = int(h5py.__version__.split('.')[0])

        for i, col in enumerate(dt.dataColumns):
            coldata = gotukb['h5key/{}'.format(colnames[i])]

            if col.name in ('1-0.0', '2-0.0'):

                # h5py 3.x and newer will return variable
                # length UTF8 strings as bytes objects,
                # but h5py 2.x will return str objects.
                if h5ver >= 3:
                    coldata = list(coldata.asstr())

                expcol = exporting.defaultDateTimeFormat(
                    dt, col, exp[col.name])
            else:
                expcol = exp[col.name]

            assert np.all(np.asarray(coldata) == expcol)


def test_exportData_HDF_subject_ordering():

    custom.registerBuiltIns()

    with tempdir():
        data   = np.random.randint(1, 10, (5, 20))
        dtable = gen_DataTable(data)
        subjs  = np.arange(20, 0, -1)
        dtable = dtable.subtable(rows=subjs)

        # pandas style + no chunking
        exporting.exportData(dtable,
                             'data.h5',
                             fileFormat='hdf5')
        got  = pd.read_hdf('data.h5')
        assert np.all(got.index == subjs)

        # pandas style + chunking
        exporting.exportData(dtable,
                             'data.h5',
                             fileFormat='hdf5',
                             numRows=2)
        got  = pd.read_hdf('data.h5')
        assert np.all(got.index == subjs)

        # funpack style
        exporting.exportData(dtable,
                             'data.h5',
                             style='funpack',
                             fileFormat='hdf5')

        got = exporting_hdf5.importFunpackStyle(
            'data.h5', 'eid', key='funpack')

        assert np.all(got.index == subjs)



def test_compound():

    data = tw.dedent("""
    "eid","1-0.0","2-0.0"
    1,"1,2,3",8
    2,"4,5,6",8
    3,"7,8,9",8
    """).strip()

    exp = tw.dedent("""
    eid\t1-0.0\t2-0.0
    1\t1.0,2.0,3.0\t8.0
    2\t4.0,5.0,6.0\t8.0
    3\t7.0,8.0,9.0\t8.0
    """).strip()


    def parseCompound(v):
        return np.fromstring(v, sep=',')

    with tempdir():

        with open('data.txt', 'wt') as f:
            f.write(data)

        vartable, proctable, cattable, _ = gen_tables(
            [1, 2], vtypes={1:'compound'})
        finfo = fileinfo.FileInfo('data.txt')
        dt, _ = importing.importData(finfo, vartable, proctable, cattable)

        dt[:, '1-0.0'] = dt[:, '1-0.0'].apply(parseCompound)

        for col in dt.dataColumns:
            exporting.formatColumn(col,
                                   dt,
                                   'default',
                                   'default',
                                   {1 : 'compound'})

        exporting.exportData(dt,
                             'out.txt',
                             fileFormat='tsv')


        with open('out.txt', 'rt') as f:
            got = f.read().strip()

        assert got == exp


def test_exporting_id_column():
    data = tw.dedent("""
    my_id,col1
    1,11
    2,12
    3,13
    """).strip()

    custom.registerBuiltIns()

    with tempdir():
        with open('data.txt', 'wt') as f:
            f.write(data)

        vartable, proctable, cattable, _ = gen_tables([1], datafiles=['data.txt'])
        finfo = fileinfo.FileInfo('data.txt')
        dt, _ = importing.importData(finfo, vartable, proctable, cattable)

        exporting.exportData(dt,
                             'out.txt',
                             fileFormat='tsv')
        got = pd.read_csv('out.txt', delimiter='\t', index_col=0)
        assert got.index.name == 'my_id'


def test_exporting_no_data():

    data = tw.dedent("""
    eid,1-0.0,2-0.0
    1,11,21
    2,12,22
    3,13,23
    4,14,24
    5,15,25
    """)

    custom.registerBuiltIns()

    with tempdir():
        with open('data.txt', 'wt') as f:
            f.write(data)

        vartable, proctable, cattable, _ = gen_tables([1])
        finfo = fileinfo.FileInfo('data.txt')
        dt, _ = importing.importData(
            finfo, vartable, proctable, cattable, variables=[3])

        with pytest.raises(RuntimeError):
            exporting.exportData(dt,
                                 'out.txt',
                                 fileFormat='tsv')
        assert not op.exists('out.txt')


def test_exportTSV_parallel():

    ncols  = 100
    nsubjs = 50000

    data       = np.random.randint(1, 100, (nsubjs, ncols + 1))
    data[:, 0] = np.arange(1, nsubjs + 1)
    colnames   = ['eid'] + ['{}-0.0'.format(i) for i in range(1, ncols + 1)]

    custom.registerBuiltIns()

    with tempdir():
        with open('data.txt', 'wt') as f:
            f.write('\t'.join(colnames) + '\n')
            np.savetxt(f, data, fmt='%i', delimiter='\t')

        vartable, proctable, cattable, _ = gen_tables(range(1, ncols + 1))
        finfo = fileinfo.FileInfo('data.txt')
        dt, _ = importing.importData(
            finfo, vartable, proctable, cattable)

        # in one go, single process
        exporting_tsv.exportTSV(dt,
                                'out1.tsv',
                                numRows=100000)

        # chunked, single process
        exporting_tsv.exportTSV(dt,
                                'out2.tsv',
                                numRows=8767)

        # chunked, multiprocess
        with mp.Pool(8) as pool:
            with mock.patch.object(dt, 'pool', return_value=pool):
                exporting_tsv.exportTSV(dt,
                                        'out3.tsv',
                                        numRows=5675)

        exp  = dt[:, :]
        got1 = pd.read_csv('out1.tsv', sep='\t', index_col=0)
        got2 = pd.read_csv('out2.tsv', sep='\t', index_col=0)
        got3 = pd.read_csv('out3.tsv', sep='\t', index_col=0)

        def eq(df):
            assert np.all(df.eq(exp))
            assert np.all(df.columns == exp.columns)
            assert np.all(df.index   == exp.index)

        eq(got1)
        eq(got2)
        eq(got3)


def test_exportCSV_escapeNewlines():

    data = tw.dedent("""
    eid,1-0.0,2-0.0
    1,"a b c
    d e f",abc
    2,one two three,"four five
    six seven"
    3,"x y	z","a b

    cde"
    """).strip()

    exp = tw.dedent(r"""
    eid,1-0.0,2-0.0
    1,a b c\nd e f,abc
    2,one two three,four five\nsix seven
    3,x y\tz,a b\n\ncde
    """).strip()

    with tempdir():

        with open('in.csv', 'wt') as f:
            f.write(data)

        vartable, proctable, cattable, _ = gen_tables(
            [1, 2], vtypes={1 : 'text', 2 : 'text'})
        finfo = fileinfo.FileInfo('in.csv')
        dt, _ = importing.importData(
            finfo, vartable, proctable, cattable)

        exporting_tsv.exportCSV(dt, 'out.csv', sep=',', escapeNewlines=True)

        with open('out.csv', 'rt') as f:
            got = f.read()

        assert got.strip() == exp



#!/usr/bin/env python
#
# test_processing_functions.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

import string
import random
import textwrap as tw
import itertools as it
import multiprocessing as mp

import io

from unittest import mock

import numpy  as np
import pandas as pd

from pandas.testing import assert_frame_equal

import funpack.importing            as importing
import funpack.fileinfo             as fileinfo
import funpack.hierarchy            as hierarchy
import funpack.util                 as util
import funpack.processing_functions as pfns

from . import (gen_DataTable,
               gen_DataTableFromDataFrame,
               tempdir,
               patch_logging,
               gen_tables)


@patch_logging
def test_removeIfSparse():

    sparse = np.random.random(100)
    good   = np.random.random(100)

    sparse[:50] = np.nan

    dtable = gen_DataTable([good, sparse])
    remove = pfns.removeIfSparse(dtable, [1, 2], minpres=60)
    remove = [r.name for r in remove]
    assert remove == ['2-0.0']

    dtable = gen_DataTable([good, sparse])
    remove = pfns.removeIfSparse(dtable, [1, 2], minpres=40)
    remove = [r.name for r in remove]
    assert remove == []


@patch_logging
def test_removeIfSparse_typed():

    mincat = np.zeros(100)
    good   = np.zeros(100)
    maxcat = np.zeros(100)

    mincat[:5]  = 1
    maxcat[:95] = 1
    good[  :50] = 1

    dtable = gen_DataTable([mincat, good, maxcat])
    dtable.vartable.loc[1, 'Type'] = util.CTYPES.categorical_single
    dtable.vartable.loc[2, 'Type'] = util.CTYPES.categorical_single
    dtable.vartable.loc[3, 'Type'] = util.CTYPES.categorical_single
    remove = pfns.removeIfSparse(dtable, [1, 2, 3], mincat=10, maxcat=90)
    remove = [r.name for r in remove]
    assert remove == ['1-0.0', '3-0.0']

    dtable.vartable.loc[1, 'Type'] = util.CTYPES.unknown
    dtable.vartable.loc[2, 'Type'] = util.CTYPES.unknown
    dtable.vartable.loc[3, 'Type'] = util.CTYPES.unknown
    remove = pfns.removeIfSparse(dtable, [1, 2, 3], mincat=10, maxcat=90)
    remove = [r.name for r in remove]
    assert remove == []

    remove = pfns.removeIfSparse(dtable, [1, 2, 3], mincat=10, maxcat=90,
                                 ignoreType=True)
    remove = [r.name for r in remove]
    assert remove == ['1-0.0', '3-0.0']


@patch_logging
def test_removeIfRedundant():

    series1 = np.sin(np.linspace(0, np.pi * 6, 100))
    series2 = series1 + np.random.random(100)
    series3 = ['abc'] * 100
    corr    = pd.Series(series1).corr(pd.Series(series2))

    dtable = gen_DataTable([series1, series2, series3])
    remove = pfns.removeIfRedundant(dtable, [1, 2, 3], corr * 0.9)
    remove = [r.name for r in remove]
    assert remove == ['2-0.0']

    dtable = gen_DataTable([series1, series2, series3])
    remove = pfns.removeIfRedundant(dtable, [1, 2, 3], corr * 1.1)
    remove = [r.name for r in remove]
    assert remove == []


@patch_logging
def test_removeIfRedundant_nathres():

    size    = 50
    series1 = np.sin(np.linspace(0, np.pi * 6, size))
    series2 = series1 + np.random.random(size)

    # insert some missing values, making
    # sure there are more missing values
    # in series2, and the missingness will
    # be positively correlated
    s1miss = np.random.choice(
        np.arange(size, dtype=np.int), 10, replace=False)
    s2miss = list(s1miss)
    while len(s2miss) < 13:
        idx = np.random.randint(0, size, 1)
        if idx not in s1miss:
            s2miss.append(idx)
    s2miss = np.array(s2miss, dtype=np.int)

    series1[s1miss] = np.nan
    series2[s2miss] = np.nan

    corr   = pd.Series(series1).corr(pd.Series(series2))
    nacorr = np.corrcoef(np.isnan(series1), np.isnan(series2))[0, 1]

    vids   = [1, 2]
    dtable = gen_DataTable([series1, series2])

    def asrt(res, exp):
        assert [int(c.name.split('-')[0]) for c in res] == exp

    asrt(pfns.removeIfRedundant(dtable, vids, corr * 1.01), [])
    asrt(pfns.removeIfRedundant(dtable, vids, corr * 0.99), [2])

    # both present and missing values must
    # be above the threshold for the column
    # to be considered redundant
    asrt(pfns.removeIfRedundant(dtable, vids, corr * 0.99, nacorr * 0.99), [2])
    asrt(pfns.removeIfRedundant(dtable, vids, corr * 1.01, nacorr * 0.99), [])
    asrt(pfns.removeIfRedundant(dtable, vids, corr * 0.99, nacorr * 1.01), [])
    asrt(pfns.removeIfRedundant(dtable, vids, corr * 1.01, nacorr * 1.01), [])

    # the column with more missing values
    # should be the one flagged as redundant
    dtable = gen_DataTable([series2, series1])
    asrt(pfns.removeIfRedundant(dtable, vids, corr * 1.01)               , [])
    asrt(pfns.removeIfRedundant(dtable, vids, corr * 0.99)               , [1])
    asrt(pfns.removeIfRedundant(dtable, vids, corr * 0.99, nacorr * 0.99), [1])
    asrt(pfns.removeIfRedundant(dtable, vids, corr * 1.01, nacorr * 0.99), [])
    asrt(pfns.removeIfRedundant(dtable, vids, corr * 0.99, nacorr * 1.01), [])
    asrt(pfns.removeIfRedundant(dtable, vids, corr * 1.01, nacorr * 1.01), [])




@patch_logging
def test_removeIfRedundant_parallel():

    for _ in range(10):
        npts, ncols      = 50, 500
        hcols            = ncols // 2
        seed             = np.random.randint(1, 10000000)

        np.random.seed(seed)

        data             = np.random.randn(npts, ncols)
        data[:, :hcols] += data[:, hcols:]
        naprops          = np.random.random(ncols) * 0.75

        # make sure some cols are all present or all missing
        naprops[:5]  = 1
        naprops[-5:] = 0

        namask           = np.random.random((npts, ncols)) < naprops
        data[namask]     = np.nan

        df = pd.DataFrame(data)

        nacorr    = np.corrcoef(namask.T)
        dcorr     = df.corr().to_numpy()
        nathres   = np.abs(nacorr[~np.isclose(nacorr, 1)]).mean()
        corrthres = np.abs(dcorr[ ~np.isclose(dcorr,  1)]).mean()

        pardt  = gen_DataTableFromDataFrame(df, njobs=8)
        seqdt  = gen_DataTableFromDataFrame(df)
        vids   = range(1, ncols + 1)

        parrem = pfns.removeIfRedundant(pardt, vids, corrthres, nathres, pairwise=True)
        seqrem = pfns.removeIfRedundant(seqdt, vids, corrthres, nathres, pairwise=True)

        parrem = sorted(parrem, key=lambda c: c.vid)
        seqrem = sorted(seqrem, key=lambda c: c.vid)

        assert parrem == seqrem


@patch_logging
def test_removeIfRedundant_skipUnknowns():
    series1 = np.sin(np.linspace(0, np.pi * 6, 100))
    series2 = series1 + np.random.random(100)
    series3 = series2

    corr    = pd.Series(series1).corr(pd.Series(series2))
    dtable  = gen_DataTable([series1, series2, series3])
    # 1 will be in uncategorised
    dtable.cattable['some category'] = [2, 3]
    remove = pfns.removeIfRedundant(dtable, [1, 2, 3], corr * 0.9)
    remove = [r.name for r in remove]
    assert remove == ['2-0.0', '3-0.0']

    remove = pfns.removeIfRedundant(dtable, [1, 2, 3], corr * 0.9,
                                    skipUnknowns=True)
    remove = [r.name for r in remove]
    assert remove == []


@patch_logging
def test_binariseCategorical():

    # internally, funpack stores integers as float32,
    # so it can use nan to represent missing. This
    # affects column name generation (e.g. 1-0.1.0,
    # rather than 1-0.1)
    data = np.random.randint(1, 10, (50, 14)).astype(np.float32)
    data[:, 0] = np.arange(1, 51)
    cols = ['eid',
            '1-0.0', '1-1.0',
            '2-0.0', '2-1.0', '2-2.0',
            '3-0.0',
            '4-0.0', '4-0.1', '4-0.2',
            '5-0.0', '5-0.1', '5-1.0', '5-1.1']

    vids  = list(range(1, 6))
    didxs = list(range(1, 14))

    with tempdir():
        np.savetxt('data.txt', data, fmt='%i', delimiter=',', header=','.join(cols))
        vartable, proctable, cattable = gen_tables(range(1, 13))[:3]
        finfo = fileinfo.FileInfo('data.txt')
        dt, _ = importing.importData(finfo, vartable, proctable, cattable)

        remove, add, addvids, kwargs = pfns.binariseCategorical(
            dt, vids,
            acrossVisits=False,
            acrossInstances=False,
            nameFormat='{vid}-{visit}.{instance}.{value}')

        names = [cols[i] for i in didxs]
        assert [r.name for r in remove] == names

        uniqs = [(i, np.unique(data[:, i])) for i in didxs]
        offset = 0

        for didx, duniqs in uniqs:

            for u in duniqs:

                exp = data[:, didx] == u
                i = offset
                offset += 1

                assert addvids[i] == int(cols[didx].split('-')[0])
                assert add[i].name == '{}.{}'.format(cols[didx], u)
                assert (add[i] == exp).all()
                assert kwargs[i]['metadata'] == u

        remove, add, addvids, kwargs = pfns.binariseCategorical(
            dt, vids,
            acrossVisits=True,
            acrossInstances=False,
            nameFormat='{vid}.{instance}.{value}')

        names = []
        offset = 0
        for vid in vids:
            for instance in dt.instances(vid):
                cols = [c.name for c in dt.columns(vid, instance=instance)]
                uniqs = sorted(np.unique(dt[:, cols]))
                names.extend(cols)

                for u in uniqs:
                    exp = np.any(dt[:, cols] == u, axis=1)

                    i = offset
                    offset += 1
                    got = add[i]
                    assert addvids[i] == vid
                    assert got.name == '{}.{}.{}'.format(vid, instance, u)
                    assert (got == exp).all()
                    assert kwargs[i]['metadata'] == u
        assert names == [r.name for r in remove]

        remove, add, addvids, kwargs = pfns.binariseCategorical(
            dt, vids,
            acrossVisits=False,
            acrossInstances=True,
            nameFormat='{vid}.{visit}.{value}')

        names = []
        offset = 0
        for vid in vids:
            for visit in dt.visits(vid):
                cols = [c.name for c in dt.columns(vid, visit=visit)]
                uniqs = sorted(np.unique(dt[:, cols]))
                names.extend(cols)

                for u in uniqs:
                    exp = np.any(dt[:, cols] == u, axis=1)

                    i = offset
                    offset += 1
                    got = add[i]
                    assert addvids[i] == vid
                    assert got.name == '{}.{}.{}'.format(vid, visit, u)
                    assert (got == exp).all()
                    assert kwargs[i]['metadata'] == u
        assert names == [r.name for r in remove]

        remove, add, addvids, kwargs = pfns.binariseCategorical(
            dt, vids,
            acrossVisits=True,
            acrossInstances=True,
            nameFormat='{vid}.{value}')

        names = []
        offset = 0
        for vid in vids:
            cols = [c.name for c in dt.columns(vid)]
            uniqs = sorted(np.unique(dt[:, cols]))
            names.extend(cols)

            for u in uniqs:
                exp = np.any(dt[:, cols] == u, axis=1)

                i = offset
                offset += 1
                got = add[i]
                assert addvids[i] == vid
                assert got.name == '{}.{}'.format(vid, u)
                assert (got == exp).all()
                assert kwargs[i]['metadata'] == u
        assert names == [r.name for r in remove]


@patch_logging
def test_binariseCateorical_no_replace():

    data = np.random.randint(1, 10, (50, 3)).astype(np.float32)
    data[:, 0] = np.arange(1, 51)
    cols = ['eid', '1-0.0', '2-0.0']
    vids  = [1, 2]

    with tempdir():
        np.savetxt('data.txt', data, delimiter=',', header=','.join(cols))
        vartable, proctable, cattable = gen_tables(vids)[:3]
        finfo = fileinfo.FileInfo('data.txt')
        dt, _ = importing.importData(finfo, vartable, proctable, cattable)

        remove, add, addvids, kwargs = pfns.binariseCategorical(
            dt, [1], replace=False, nameFormat='{vid}.{value}')

        uniq = np.unique(data[:, 1])

        assert len(remove)  == 0
        assert len(add)     == len(uniq)
        assert len(addvids) == len(uniq)
        meta = [kw['metadata'] for kw in kwargs]
        assert sorted(uniq) == sorted(meta)

        assert sorted([c.name for c in add]) == \
               sorted(['1.{}'.format(u) for u in uniq])
        assert np.all([v == 1 for v in addvids])


@patch_logging
def test_binariseCategorical_nonnumeric():

    data = [random.choice(string.ascii_letters[:8]) for i in range(40)]
    data = [data[:20], data[20:]]

    with tempdir():
        with open('data.txt', 'wt') as f:
            f.write('eid, 1-0.0, 1-1.0\n')
            for i, (v1, v2) in enumerate(zip(*data)):
                f.write('{}, {}, {}\n'.format(i + 1, v1, v2))

        vartable, proctable, cattable = gen_tables(
            [1], vtypes={1 : 'text'})[:3]
        finfo = fileinfo.FileInfo('data.txt')
        dt, _ = importing.importData(finfo, vartable, proctable, cattable)

        remove, add, addvids, kwargs = pfns.binariseCategorical(
            dt, [1],
            acrossVisits=True,
            acrossInstances=True,
            nameFormat='{vid}.{value}')

        unique   = set(data[0] + data[1])
        remnames = [r.name for r in remove]
        addnames = [a.name for a in add]

        assert '1-0.0' in remnames and '1-1.0' in remnames
        assert len(addvids) == len(add)
        assert all([v == 1 for v in addvids])
        meta = [kw['metadata'] for kw in kwargs]
        assert sorted(unique) == sorted(meta)

        for u in unique:
            name = '1.{}'.format(u)
            assert name in addnames

            series = add[addnames.index(name)]
            mask = [u == data[0][i] or u == data[1][i] for i in range(20)]

            assert (series == mask).all()


@patch_logging
def test_binariseCategorical_take():
    data = tw.dedent("""
    1,A,B,C,D,E,1,2,3,4,5
    2,A,,,,,6,,,,
    3,B,D,E,,,7,8,9,,
    4,D,E,,,,10,11,,,
    5,C,E,,,,12,13,,,
    6,A,C,D,,,14,15,16,,
    7,B,D,,,,17,18,,,
    8,C,,,,,19,,,,
    9,,,,,,,,,,
    10,B,E,,,,20,21,,,
    """).strip()

    # across all columns
    cols1 = ['eid'] + \
            ['1-0.1', '1-0.2', '1-0.3', '1-0.4', '1-0.5'] + \
            ['2-0.1', '2-0.2', '2-0.3', '2-0.4', '2-0.5']
    cols1 = ','.join(cols1)
    exp1  = tw.dedent("""
    eid,1-A,1-B,1-C,1-D,1-E
    1,1,2,3,4,5
    2,6,,,,
    3,,7,,8,9
    4,,,,10,11
    5,,,12,,13
    6,14,,15,16,
    7,,17,,18,
    8,,,19,,
    9,,,,,
    10,,20,,,21
    """).strip()

    # across instances (within visit)
    cols2 = ['eid'] + \
            ['1-0.0', '1-0.1', '1-1.0', '1-1.1', '1-1.2'] + \
            ['2-0.0', '2-0.1', '2-1.0', '2-1.1', '2-1.2']
    cols2 = ','.join(cols2)
    exp2  = tw.dedent("""
    eid,1-0.A,1-0.B,1-0.C,1-0.D,1-0.E,1-1.C,1-1.D,1-1.E
    1,1,2,,,,3,4,5
    2,6,,,,,,,
    3,,7,,8,,,,9
    4,,,,10,11,,,
    5,,,12,,13,,,
    6,14,,15,,,,16,
    7,,17,,18,,,,
    8,,,19,,,,,
    9,,,,,,,,
    10,,20,,,21,,,
    """).strip()

    # across visits (within instance)
    cols3 = ['eid'] + \
            ['1-0.0', '1-1.0', '1-0.1', '1-1.1', '1-2.1'] + \
            ['2-0.0', '2-1.0', '2-0.1', '2-1.1', '2-2.1']
    cols3 = ','.join(cols3)
    exp3  = tw.dedent("""
    eid,1-A.0,1-B.0,1-C.0,1-D.0,1-E.0,1-C.1,1-D.1,1-E.1
    1,1,2,,,,3,4,5
    2,6,,,,,,,
    3,,7,,8,,,,9
    4,,,,10,11,,,
    5,,,12,,13,,,
    6,14,,15,,,,16,
    7,,17,,18,,,,
    8,,,19,,,,,
    9,,,,,,,,
    10,,20,,,21,,,
    """).strip()

    # within each column
    # across all columns
    cols4 = ['eid'] + \
            ['1-0.0', '1-0.1', '1-0.2', '1-0.3', '1-0.4'] + \
            ['2-0.0', '2-0.1', '2-0.2', '2-0.3', '2-0.4']
    cols4 = ','.join(cols4)
    exp4 = tw.dedent("""
    eid,1-0.0_A,1-0.0_B,1-0.0_C,1-0.0_D,1-0.1_B,1-0.1_C,1-0.1_D,1-0.1_E,1-0.2_C,1-0.2_D,1-0.2_E,1-0.3_D,1-0.4_E
    1,1,,,,2,,,,3,,,4,5
    2,6,,,,,,,,,,,,
    3,,7,,,,,8,,,,9,,
    4,,,,10,,,,11,,,,,
    5,,,12,,,,,13,,,,,
    6,14,,,,,15,,,,16,,,
    7,,17,,,,,18,,,,,,
    8,,,19,,,,,,,,,,
    9,,,,,,,,,,,,,
    10,,20,,,,,,21,,,,,
    """).strip()

    tests = [(cols1, exp1, True, True),
             (cols2, exp2, False, True),
             (cols3, exp3, True, False),
             (cols4, exp4, False, False),
    ]

    with tempdir():
        for cols, exp, av, ai in tests:
            with open('data.txt', 'wt') as f:
                f.write(cols + '\n')
                f.write(data)

            vartable, proctable, cattable = gen_tables(
                [1, 2], vtypes={1 : 'text'})[:3]
            finfo = fileinfo.FileInfo('data.txt')
            dt, _ = importing.importData(finfo, vartable, proctable, cattable)

            remove, add, addvids, kwargs = pfns.binariseCategorical(
                dt, [1],
                acrossVisits=av,
                acrossInstances=ai, take=2)

            exp = pd.read_csv(io.StringIO(exp), index_col=0)
            got = pd.DataFrame({a.name : a for a in add})
            assert_frame_equal(got, exp, check_dtype=False, check_like=False)



def test_binariseCategorical_missing_vars():

    data = np.random.randint(1, 10, (50, 5))
    data[:, 0] = np.arange(1, 51)
    cols = ['eid', '1-0.0', '2-0.0', '3-0.0', '4-0.0']
    vids  = [1, 2, 3, 4]

    nuniq = {i :  len(np.unique(data[:, i])) for i in [1, 2, 3, 4]}

    with tempdir():
        np.savetxt('data.txt', data, delimiter=',', header=','.join(cols))
        vartable, proctable, cattable = gen_tables(vids)[:3]
        finfo = fileinfo.FileInfo('data.txt')
        dt, _ = importing.importData(finfo, vartable, proctable, cattable)

        series = pfns.binariseCategorical(dt, [1], take=[3])[1]
        assert len(series) == nuniq[1]

        series = pfns.binariseCategorical(dt, [1, 2], take=[3, 4])[1]
        assert len(series) == nuniq[1] + nuniq[2]

        series = pfns.binariseCategorical(dt, [1], take=[5])[1]
        assert len(series) == 0

        series = pfns.binariseCategorical(dt, [1, 2], take=[3, 5])[1]
        assert len(series) == nuniq[1]

        series = pfns.binariseCategorical(dt, [1, 5], take=[3, 6])[1]
        assert len(series) == nuniq[1]

        series = pfns.binariseCategorical(dt, [5])[1]
        assert len(series) == 0

        series = pfns.binariseCategorical(dt, [5], take=[6])[1]
        assert len(series) == 0

        series = pfns.binariseCategorical(dt, [5], take=[6])[1]
        assert len(series) == 0


# fsl/funpack!100
def test_binariseCategorical_unique_colnames():

    df = pd.DataFrame({'1-0.0' : pd.Series(['1', '01', '2', '002']),
                       'eid'   : [1, 2, 3, 4]}).set_index('eid')

    with tempdir():
        df.to_csv('data.csv')
        vartable, proctable, cattable = gen_tables(
            [1], vtypes={1 : 'text'})[:3]
        finfo = fileinfo.FileInfo('data.csv')
        dt, _ = importing.importData(finfo, vartable, proctable, cattable)

        series = pfns.binariseCategorical(dt, [1], acrossVisits=True)[1]
        assert len(series) == 4
        names = sorted([s.name for s in series])
        assert names == sorted(['1-1', '1-01', '1-2', '1-002'])

        got = pd.DataFrame({s.name : s for s in series})
        assert (got['1-1']   == [1, 0, 0, 0]).all()
        assert (got['1-01']  == [0, 1, 0, 0]).all()
        assert (got['1-2']   == [0, 0, 1, 0]).all()
        assert (got['1-002'] == [0, 0, 0, 1]).all()


@patch_logging
def test_expandCompound():

    data = []

    for i in range(20):
        rlen = np.random.randint(1, 20)
        row = np.random.randint(1, 100, rlen)
        data.append(row)

    exp = np.full((len(data), max(map(len, data))), np.nan)

    for i in range(len(data)):
        exp[i, :len(data[i])] = data[i]

    with tempdir():
        with open('data.txt', 'wt') as f:
            f.write('eid\t1-0.0\n')
            for i in range(len(data)):
                f.write(str(i + 1) + '\t' + ','.join(map(str, data[i])))
                f.write('\n')

        vartable, proctable, cattable = gen_tables(
            [1], vtypes={1 : 'text'})[:3]
        finfo = fileinfo.FileInfo('data.txt')
        dt, _ = importing.importData(finfo, vartable, proctable, cattable)

        dt[:, '1-0.0'] = dt[:, '1-0.0'].apply(
            lambda v: np.fromstring(v, sep=','))

        remove, add, addvids = pfns.expandCompound(dt, [1])

        assert [r.name for r in remove] == ['1-0.0']
        assert len(add) == max(map(len, data))
        assert len(addvids) == len(add)
        assert all([v == 1 for v in addvids])

        for col in range(exp.shape[1]):
            expcol = exp[:, col]
            gotcol = add[col].values

            expna = np.isnan(expcol)
            gotna = np.isnan(gotcol)

            assert np.all(        expna  ==         gotna)
            assert np.all(expcol[~expna] == gotcol[~gotna])



@patch_logging
def test_createDiagnosisColumns():

    maxdiags = 8
    nsubjs   = 10
    maincols = [f'41270-0.{i}' for i in range(1, 1 + maxdiags)]
    primcols = [f'41202-0.{i}' for i in range(1, 1 + maxdiags // 2)]
    seccols  = [f'41204-0.{i}' for i in range(1, 1 + maxdiags // 2)]
    allcols  = maincols + primcols + seccols
    h        = hierarchy.loadHierarchyFile(hierarchy.getHierarchyFilePath(coding=19))
    df       = pd.DataFrame(index=np.arange(1, nsubjs + 1), columns=allcols)
    df.index.name = 'eid'

    allcodes  = set()
    primcodes = set()
    seccodes  = set()

    for subj in range(1, nsubjs + 1):
        numdiags  = np.random.randint(1, maxdiags + 1, 1)[0]
        maindiags = np.random.choice(np.arange(1, len(h.codings) + 1), numdiags)
        maindiags = [h.codings[d] for d in maindiags]
        primdiags = maindiags[:max(1, numdiags // 2)]
        secdiags  = maindiags[ max(1, numdiags // 2):]

        allcodes .update(maindiags)
        primcodes.update(primdiags)
        seccodes .update(secdiags)

        maindiags = maindiags + [np.nan for _ in range(maxdiags      - len(maindiags))]
        primdiags = primdiags + [np.nan for _ in range(maxdiags // 2 - len(primdiags))]
        secdiags  = secdiags  + [np.nan for _ in range(maxdiags // 2 - len(secdiags))]

        df.loc[subj, maincols] = maindiags
        df.loc[subj, primcols] = primdiags
        df.loc[subj, seccols]  = secdiags

    def check(remove, add, addvids):
        # check that there are new primary and
        # secondary columns for each code
        primadded = [s.name for s in add if s.name.startswith('41202')]
        secadded  = [s.name for s in add if s.name.startswith('41204')]
        primadded = [s.split('-', 1)[1].split('.')[0] for s in primadded]
        secadded  = [s.split('-', 1)[1].split('.')[0] for s in secadded]

        assert sorted(allcodes) == sorted(primadded)
        assert sorted(allcodes) == sorted(secadded)

        for col, vid in zip(add, addvids):
            if col.name.startswith('41202'): assert vid == 41202
            else:                            assert vid == 41204

        # check that primary/secondary code
        # columns are flagged for removal
        assert sorted(primcols + seccols) == [r.name for r in remove]

        # check that primary/secondary
        # codes were flagged correctly
        primadded = {}
        secadded  = {}
        for col in add:
            code, ptype = col.name.split('-', 1)[1].split('.')

            if ptype == 'primary': primadded[code] = col
            else:                  secadded[ code] = col

        for subj in df.index:
            primdiags = df.loc[subj, primcols].dropna()
            secdiags  = df.loc[subj, seccols] .dropna()

            for d in primdiags:
                assert primadded[d][subj] == 1
                assert secadded[ d][subj] == 0
            for d in secdiags:
                assert primadded[d][subj] == 0
                assert secadded[ d][subj] == 1

    with tempdir():
        df.to_csv('data.csv')

        vartable, proctable, cattable = gen_tables(
            [41270, 41202, 41204], vtypes={41270 : 'text',
                                           41202 : 'text',
                                           41204 : 'text'})[:3]
        finfo = fileinfo.FileInfo('data.csv')

        # test from raw data
        dt, _ = importing.importData(finfo, vartable, proctable, cattable)
        remove, add, addvids, _ = pfns.createDiagnosisColumns(dt, [41270], 41202, 41204)
        check(remove, add, addvids)

        # test from data that has been
        # passed through binariseCategorical
        dt, _ = importing.importData(finfo, vartable, proctable, cattable)
        remove, add, addvids, kwargs = pfns.binariseCategorical(dt, [41270])
        if len(remove) > 0: dt.removeColumns(remove)
        if len(add)    > 0: dt.addColumns(add, addvids, kwargs)
        remove, add, addvids, _ = pfns.createDiagnosisColumns(dt, [41270], 41202, 41204, binarised=True)
        check(remove, add, addvids)



@patch_logging
def test_removeField():

    v1 = np.random.random(100)
    v2 = np.random.random(100)
    v3 = np.random.random(100)

    dtable = gen_DataTable([v1, v2, v3])
    cols   = [None] + dtable.dataColumns

    for vids in it.permutations([1, 2, 3], 2):
        assert pfns.removeField(dtable, vids) == [cols[v] for v in vids]
    for vids in it.permutations([1, 2, 3], 2):
        assert pfns.removeField(dtable, vids) == [cols[v] for v in vids]
    for vids in it.permutations([1, 2, 3], 3):
        assert pfns.removeField(dtable, vids) == [cols[v] for v in vids]



#!/usr/bin/env python
#
# exporting_tsv.py - Export data to a TSV file.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :func:`exportTSV` and :func:`exportCSV` functions,
which export the data contained in a :class:`.DataTable` to a TSV or CSV file.
"""


import functools as ft
import os.path   as op
import              os
import              logging

import numpy            as np
import pandas.api.types as pdtypes

from . import util
from . import custom
from . import exporting


log = logging.getLogger(__name__)


NUM_ROWS = 10000
"""Default number of rows to export at a time by :func:`exportTSV` - the
default value for its ``numRows`` argument.
"""


@custom.exporter('csv')
def exportCSV(dtable, outfile, **kwargs):
    """Export data to a CSV-style file.

    This function is identical to the :func:`exportTSV`, except that the
    default value for the `sep`` argument is a ``','`` instead of a ``'\t'``.
    """

    sep = kwargs.pop('sep', None)
    if sep is None:
        sep = ','

    exportTSV(dtable, outfile, sep=sep, **kwargs)


@custom.exporter('tsv')
def exportTSV(dtable,
              outfile,
              sep=None,
              missingValues=None,
              escapeNewlines=False,
              numRows=None,
              dropNaRows=False,
              dateFormat=None,
              timeFormat=None,
              formatters=None,
              **kwargs):
    """Export data to a TSV-style file.

    This may be parallelised by row - chunks of ``numRows`` rows will be
    saved to separate temporary output files in parallel, and then concatenated
    afterwards to produce the final output file.

    :arg dtable:         :class:`.DataTable` containing the data

    :arg outfile:        File to output to

    :arg sep:            Separator character to use. Defaults to ``'\\t'``

    :arg missingValues:  String to use for missing/NA values. Defaults to the
                         empty string.


    :arg escapeNewlines: If ``True``, all string/object types are escaped
                         using ``shlex.quote``.

    :arg numRows:        Number of rows to write at a time. Defaults to
                         :attr:`NUM_ROWS`.

    :arg dropNaRows:     If ``True``, rows which do not contain data for any
                         columns are not exported.

    :arg dateFormat:     Name of formatter to use for date columns.

    :arg timeFormat:     Name of formatter to use for time columns.

    :arg formatters:     Dict of ``{ [vid|column] : formatter }`` mappings,
                         specifying custom formatters to use for specific
                         variables.
    """

    if sep           is None: sep           = '\t'
    if missingValues is None: missingValues = ''
    if numRows       is None: numRows       = NUM_ROWS
    if formatters    is None: formatters    = {}

    # We're going to output each chunk of
    # subjects to a separate file (in
    # parallel), and then cat the files
    # together afterwards
    index      = dtable.index
    nchunks    = int(np.ceil(len(index) / numRows))
    idxchunks  = [index[i:i + numRows] for i in
                  range(0, len(index), numRows)]
    subtables  = [dtable.subtable(rows=c) for c in idxchunks]
    outfiles   = ['{}_{}'.format(outfile, i) for i in range(nchunks)]

    # escapeNewlines is performed by the
    # exporting.escapeString formatter
    # function. We apply it to all columns
    # which are not numeric or date
    if escapeNewlines:
        for col in dtable.dataColumns:
            series = dtable[:, col.name]
            if not (pdtypes.is_numeric_dtype(       series) or
                    pdtypes.is_datetime64_any_dtype(series)):
                formatters[col.name] = 'escapeString'

    # write each chunk in parallel
    args = zip(subtables,
               outfiles,
               [True] + [False] * (nchunks - 1),
               range(nchunks))
    func = ft.partial(writeDataFrame,
                      sep=sep,
                      missingValues=missingValues,
                      dropNaRows=dropNaRows,
                      dateFormat=dateFormat,
                      timeFormat=timeFormat,
                      formatters=formatters)

    with dtable.pool() as pool:
        pool.starmap(func, args)

    # concatenate the chunks to
    # form the final output file
    if len(outfiles) == 1:
        os.rename(outfiles[0], outfile)
    else:
        util.cat(outfiles, outfile)

    # remove intermediate files
    for f in outfiles:
        if f is not None and op.exists(f):
            os.remove(f)


def writeDataFrame(dtable,
                   outfile,
                   header,
                   chunki,
                   sep,
                   missingValues,
                   dropNaRows,
                   dateFormat,
                   timeFormat,
                   formatters):
    """Writes all of the data in ``dtable`` to ``outfile``.

    Called by :func:`exportTSV` to output one chunk of data.

    :arg dtable:         :class:`.DataTable` containing the data

    :arg outfile:        File to output to

    :arg header:         If ``True``, write the header row (column names).

    :arg chunki:         Chunk index (used for logging)

    :arg sep:            Separator character to use.

    :arg missingValues:  String to use for missing/NA values.

    :arg dropNaRows:     If ``True``, rows which do not contain data for any
                         columns are not exported.

    :arg dateFormat:     Name of formatter to use for date columns.

    :arg timeFormat:     Name of formatter to use for time columns.

    :arg formatters:     Dict of ``{ [vid|column] : formatter }`` mappings,
                         specifying custom formatters to use for specific
                         variables.
    """

    for coli, col in enumerate(dtable.dataColumns):

        logmsg = 'chunk {}, col {} / {}'.format(
            chunki, coli, len(dtable.dataColumns))

        exporting.formatColumn(
            col,
            dtable,
            dateFormat=dateFormat,
            timeFormat=timeFormat,
            formatters=formatters,
            logmsg=logmsg)

    towrite = dtable[:, :]

    if dropNaRows:
        towrite.dropna(how='all', inplace=True)

    log.info('Writing %u columns and %u rows [chunk %u] to %s ...',
             len(dtable.dataColumns), len(dtable), chunki, outfile)

    if header: idcol = towrite.index.name
    else:      idcol = None

    towrite.to_csv(outfile,
                   sep=sep,
                   na_rep=missingValues,
                   header=header,
                   index_label=idcol)



#!/usr/bin/env python
#
# test_processing_core.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

import string
import random
import datetime
import itertools as it
import functools as ft

import pytest

import numpy  as np
import pandas as pd


import funpack.util                      as util
import funpack.processing_functions_core as core


def test_isSparse_minpres():

    size = 100
    actual_present    = [0,  0.01, 0.1, 0.5, 1]
    minpres_threshold = [0,  0.01, 0.1, 0.5, 1]

    for present, threshold in it.product(actual_present, minpres_threshold):

        data     = np.random.random(size)
        expected = present < threshold

        missing = int(round(size - present * size))
        missing = np.random.choice(range(size), missing, replace=False)
        data[missing] = np.nan

        data = pd.Series(data)

        absres  = core.isSparse(
            data, util.CTYPES.continuous, minpres=threshold * size)
        propres = core.isSparse(
            data, util.CTYPES.continuous, minpres=threshold, abspres=False)

        if expected:
            expcause = 'minpres'
            expval   =  size - len(missing)
        else:
            expcause = None
            expval   = None

        assert absres  == (expected, expcause, expval)
        assert propres == (expected, expcause, expval)

    # minpres should be ignored if
    # number of points in data is
    # less than or equal to it
    data = np.random.random(10)
    data[:2] = np.nan

    res = core.isSparse(pd.Series(data),
                         util.CTYPES.continuous,
                         minpres=9)
    assert res == (True, 'minpres', 8)

    res = core.isSparse(pd.Series(data),
                        util.CTYPES.continuous,
                        minpres=10)
    assert res == (True, 'minpres', 8)

    res = core.isSparse(pd.Series(data),
                        util.CTYPES.continuous,
                        minpres=11)
    assert res == (False, None, None)

    res = core.isSparse(pd.Series(data),
                        util.CTYPES.continuous,
                        minpres=100)
    assert res == (False, None, None)


def test_isSparse_minstd():

    actualstds = np.linspace(0, 2, 10)
    minstds    = np.linspace(0, 2, 10)

    size = 500

    for actualstd, minstd in it.product(actualstds, minstds):

        data      = np.random.randn(size) * actualstd
        data      = pd.Series(data)
        actualstd = data.std()
        expected  = actualstd <= minstd

        result = core.isSparse(data, util.CTYPES.continuous,
                               minstd=minstd)
        if expected:
            assert result[:2] == (expected, 'minstd')
            assert np.isclose(result[2], actualstd)


def test_isSparse_maxcat():

    size          = 20
    actualmaxcats = np.arange(1, 21, 2)
    maxcats       = np.arange(1, 21, 2)

    dtypes = [util.CTYPES.integer,
              util.CTYPES.categorical_single,
              util.CTYPES.categorical_multiple]

    # test should only be applied
    # for integer/categoricals
    data = pd.Series(np.arange(size))
    result = core.isSparse(data, util.CTYPES.continuous, maxcat=1)
    assert result == (False, None, None)

    # threshold should be ignored if
    # bigger than size of data set
    data[:] = 123
    result = core.isSparse(data, util.CTYPES.integer, maxcat=25)
    assert result == (False, None, None)

    for actualmaxcat, maxcat in it.product(actualmaxcats, maxcats):

        data = np.arange(size)
        data[:actualmaxcat] = size + 1

        data = pd.Series(data)

        expected = actualmaxcat >= maxcat

        if expected:
            expected = (expected, 'maxcat', actualmaxcat)
        else:
            expected = (expected, None, None)

        for dt in dtypes:

            maxcatprop = maxcat / len(data)

            resultabs  = core.isSparse(data, dt, maxcat=maxcat)
            resultprop = core.isSparse(data, dt, maxcat=maxcatprop,
                                       abscat=False)

            assert resultabs  == expected
            assert resultprop == expected

    # when maxcat is a proportion, test should ignore nans
    data        = np.zeros(100, dtype=np.float)
    data[:40]   = np.nan
    data[40:80] = 1
    data        = pd.Series(data)
    assert core.isSparse(data, None, maxcat=0.5, abscat=False) == \
        (True, 'maxcat', 40)
    assert core.isSparse(data, None, maxcat=0.6, abscat=False) == \
        (True, 'maxcat', 40)
    assert core.isSparse(data, None, maxcat=0.7, abscat=False) == \
        (False, None, None)


def test_isSparse_mincat():

    size          = 100
    actualmincats = np.arange(1, 21, 2)
    mincats       = np.arange(1, 21, 2)

    dtypes = [util.CTYPES.integer,
              util.CTYPES.categorical_single,
              util.CTYPES.categorical_multiple]

    # test should only be applied
    # for integer/categoricals
    data     = np.zeros(size)
    data[:5] = size + 1
    data     = pd.Series(data)
    result   = core.isSparse(data, util.CTYPES.continuous,
                             mincat=2)
    assert result == (False, None, None)

    # threshold should be ignored if
    # bigger than size of data set
    data[:] = size + 1
    result = core.isSparse(data, util.CTYPES.integer, mincat=size + 5)
    assert result == (False, None, None)

    for actualmincat, mincat in it.product(actualmincats, mincats):

        data = np.zeros(size)
        data[:actualmincat] = size + 1

        data = pd.Series(data)

        if actualmincat < mincat:
            expected = (True, 'mincat', actualmincat)
        else:
            expected = (False, None, None)

        for dt in dtypes:

            mincatprop = mincat / len(data)

            resultabs  = core.isSparse(data, dt, mincat=mincat)
            resultprop = core.isSparse(data, dt, mincat=mincatprop,
                                       abscat=False)

            assert resultabs  == expected
            assert resultprop == expected

    # when mincat is a proportion, test should ignore nans
    data        = np.zeros(100, dtype=np.float)
    data[:40]   = np.nan
    data[40:60] = 1
    data        = pd.Series(data)
    assert core.isSparse(data, None, mincat=0.1, abscat=False) == \
        (False, None, None)
    assert core.isSparse(data, None, mincat=0.2, abscat=False) == \
        (False, None, None)
    assert core.isSparse(data, None, mincat=0.3, abscat=False) == \
        (False, None, None)
    assert core.isSparse(data, None, mincat=0.4, abscat=False) == \
        (True, 'mincat', 20)

# https://git.fmrib.ox.ac.uk/fsl/funpack/-/merge_requests/81
def test_isSparse_mincat_maxcat_empty():

    data = np.full(10, np.nan, dtype=np.float32)
    data = pd.Series(data)

    result = core.isSparse(data, util.CTYPES.categorical_single, mincat=2)
    assert result == (True, 'mincat', 0)

    result = core.isSparse(data, util.CTYPES.categorical_single, maxcat=2)
    assert result == (True, 'maxcat', 0)


def test_isSparse_non_numeric():


    data     = list(np.random.randint(1, 10, 100))
    data[50] = 'abcde'
    data[51] = np.nan
    data     = pd.Series(data)
    assert core.isSparse(data, minstd=10) == (False, None, None)

    data = pd.Series(['a', 'a', 'b', 'b', 'b', np.nan, np.nan])
    assert core.isSparse(data, mincat=1) == (False, None, None)
    assert core.isSparse(data, mincat=2) == (False, None, None)
    assert core.isSparse(data, mincat=3) == (True, 'mincat', 2)
    assert core.isSparse(data, maxcat=4) == (False, None, None)
    assert core.isSparse(data, maxcat=3) == (True, 'maxcat', 3)


def test_isSparse_naval():

    data = np.random.randint(1, 10, 100)
    data = pd.Series(data)

    data.iloc[:10] = -1
    assert core.isSparse(data, minpres=89, naval=-1) == (False, None,      None)
    assert core.isSparse(data, minpres=90, naval=-1) == (False, None,      None)
    assert core.isSparse(data, minpres=91, naval=-1) == (True,  'minpres', 90)

    # nan is not missing
    data.iloc[:10] = np.nan
    assert core.isSparse(data, minpres=91, naval=-1) == (False, None, None)

    # non-numeric
    data = random.choices(string.ascii_lowercase, k=100)
    data = pd.Series(data)

    data.iloc[:10] = 'A'
    assert core.isSparse(data, minpres=89, naval='A') == (False, None,      None)
    assert core.isSparse(data, minpres=90, naval='A') == (False, None,      None)
    assert core.isSparse(data, minpres=91, naval='A') == (True,  'minpres', 90)
    data.iloc[:10] = np.nan
    assert core.isSparse(data, minpres=91, naval='A') == (False, None, None)

    # mixed type
    data.iloc[:10] = 0
    assert core.isSparse(data, minpres=91, naval=0) == (True,  'minpres', 90)


def test_isSparse_date():

    data = []
    for i in range(100):
        data.append(datetime.date(2020,
                                  np.random.randint(1, 13),
                                  np.random.randint(1, 29)))

    data = pd.Series(data)

    data.iloc[:10] = np.nan
    assert core.isSparse(data, minpres=89) == (False, None,      None)
    assert core.isSparse(data, minpres=90) == (False, None,      None)
    assert core.isSparse(data, minpres=91) == (True,  'minpres', 90)


def test_naCorrelation():
    size    = 50
    series1 = np.sin(np.linspace(0, np.pi * 6, size))
    series2 = series1 + np.random.random(size)
    series3 = series1 + np.random.random(size)

    # insert some missing values, making
    # sure there are more missing values
    # in series2, and the missingness will
    # be positively correlated
    s1miss = np.random.choice(
        np.arange(size, dtype=np.int), 10, replace=False)
    s2miss = list(s1miss)
    while len(s2miss) < 13:
        idx = np.random.randint(0, size, 1)
        if idx not in s1miss:
            s2miss.append(idx)
    s2miss = np.array(s2miss, dtype=np.int)

    series1[s1miss] = np.nan
    series2[s2miss] = np.nan

    nacorr = np.corrcoef(np.isnan(series1), np.isnan(series2))[0, 1]

    def check(df, nathres, exp):
        got = core.naCorrelation(df.isna(), nathres)
        exp = np.array(exp)
        assert np.all(np.isclose(got, exp))

    df = pd.DataFrame({0 : series1, 1 : series2})
    check(df, nacorr * 1.01, np.eye(2))
    check(df, nacorr * 0.99, [[1, 1], [1, 1]])

    # cols with no missing values
    # should always pass the test
    df = pd.DataFrame({0 : series1, 1 : series2, 2 : series3})
    check(df, nacorr * 1.01, [[1, 0, 1],
                              [0, 1, 1],
                              [1, 1, 1]])
    check(df, nacorr * 0.99, [[1, 1, 1],
                              [1, 1, 1],
                              [1, 1, 1]])


def wrapped_pairwiseRedundantColumns(df, corrthres):
    xs, ys   = np.triu_indices(len(df.columns), k=1)
    colpairs = np.vstack((xs, ys)).T
    return core.pairwiseRedundantColumns(df, colpairs, corrthres)

def test_pairwiseRedundantColumns():
    with np.errstate(divide='ignore'):
        _test_redundantColumns(wrapped_pairwiseRedundantColumns)

def test_matrixRedundantColumns():
    with np.errstate(divide='ignore'):
        _test_redundantColumns(core.matrixRedundantColumns)

def _test_redundantColumns(rcfunc):

    size    = 50
    series1 = np.sin(np.linspace(0, np.pi * 6, size))
    series2 = series1 + np.random.random(size)
    corr = pd.Series(series1).corr(pd.Series(series2))

    data = pd.DataFrame({0 : pd.Series(series1), 1 : pd.Series(series2)})

    assert np.all(rcfunc(data, corr * 1.01) == [])
    # later (naturally ordered) columns will be
    # the ones flagged as redundant, if they
    # have the same number of missing values
    assert np.all(rcfunc(data, corr * 0.99) == [(1, 0)])

    # the column with more missing values
    # should be the one flagged as redundant
    data.loc[0, 0] = np.nan
    corr = pd.Series(data[0]).corr(data[1])
    assert np.all(rcfunc(data, corr * 1.01) == [])
    assert np.all(rcfunc(data, corr * 0.99) == [(0, 1)])
    data.loc[0:2, 1] = np.nan
    corr = pd.Series(data[0]).corr(data[1])
    assert np.all(rcfunc(data, corr * 1.01) == [])
    assert np.all(rcfunc(data, corr * 0.99) == [(1, 0)])


def test_pairwiseRedundantColumns_associative():
    with np.errstate(divide='ignore'):
        _test_redundantColumns_associative(wrapped_pairwiseRedundantColumns)

def test_matrixRedundantColumns_associative():
    with np.errstate(divide='ignore'):
        _test_redundantColumns_associative(core.matrixRedundantColumns)

def _test_redundantColumns_associative(rcfunc):

    # Three columns a, b, and c, where
    # corr(a,b) > corr(a,c) > corr(b,c)
    cov = np.array([[10,  8,  4],
                    [ 8, 10,  2],
                    [ 4,  2, 10]])
    a, b, c = np.random.multivariate_normal([1, 1, 1], cov, 100).T

    # make sure that a is always selected
    # over b and b is always selected over
    # c, i.e. nans(a) < nans(b) < nans(c)
    b[-1]    = np.nan
    c[-2:]   = np.nan
    df       = pd.DataFrame(np.array([a, b, c]).T, columns=['a', 'b', 'c'])

    corrmat  = df.corr()
    abthres  = corrmat.loc['a', 'b']
    acthres  = corrmat.loc['a', 'c']
    bcthres  = corrmat.loc['b', 'c']

    colorder = 'abc'
    if True: # for colorder in it.permutations('abc', 3):
        df = df[list(colorder)]

        a = colorder.index('a')
        b = colorder.index('b')
        c = colorder.index('c')

        # threshold higher than all corrlations
        assert np.all(rcfunc(df,  abthres * 1.05) == [])
        # threshold lower than a+b, higher than others
        assert np.all(rcfunc(df, (abthres + acthres) / 2) == [(b, a)])
        # threshold lower than a+b, a+c, higher than b+c
        assert np.all(rcfunc(df, (acthres + bcthres) / 2) == [(b, a), (c, a)])
        # threshold lower than all - both b and c
        # should be redundant. Depending on column
        # order, c will be redundant w.r.t. a or
        # to b
        if a < b: exp = [(b, a), (c, a)]
        else:     exp = [(b, a), (c, b)]
        assert np.all(sorted(rcfunc(df,  bcthres * 0.95)) == sorted(exp))


# test that pairwise/matrix implementations produce same result
def test_redundantColumns_na():
    for _ in range(10):
        npts, ncols      = 100, 100
        hcols            = ncols // 2
        seed             = np.random.randint(1, 10000000)
        np.random.seed(seed)

        data             = np.random.randn(npts, ncols)
        data[:, :hcols] += data[:, hcols:]
        naprops          = np.random.random(ncols) * 0.75
        namask           = np.random.random((npts, ncols)) < naprops
        data[namask]     = np.nan

        df = pd.DataFrame(data)

        nacorr    = np.corrcoef(namask.T)
        dcorr     = df.corr().to_numpy()
        nathres   = np.abs(nacorr[~np.isclose(nacorr, 1)]).mean()
        corrthres = np.abs(dcorr[ ~np.isclose(dcorr,  1)]).mean()

        with np.errstate(divide='ignore'):
            nacorr   = core.naCorrelation(pd.isna(df), nathres)
            colpairs = np.where(np.triu(nacorr, k=1))
            colpairs = np.vstack(colpairs).T

            pairwise = core.pairwiseRedundantColumns(df, colpairs, corrthres)
            matrix   = core.matrixRedundantColumns(  df, corrthres, nathres)

        nacounts = np.isnan(data).sum(axis=0)
        corrmask = nacorr & (np.abs(dcorr) > corrthres)
        colpairs = np.where(np.triu(corrmask, k=1))
        colpairs = np.vstack(colpairs).T
        exp = {}
        for ci, cj in colpairs:
            if nacounts[ci] > nacounts[cj]: drop, keep = ci, cj
            else:                           drop, keep = cj, ci

            if drop not in exp:
                exp[drop] = keep
        exp = list(exp.items())

        assert np.all(sorted(exp)      == sorted(matrix))
        assert np.all(sorted(pairwise) == sorted(matrix))


def test_redundantColumns_constant_allna():

    seed = np.random.randint(1, 10000000)
    np.random.seed(seed)

    data       = np.random.randint(1, 100, (20, 10)).astype(np.float32)
    data[0, :] = 20
    data[1, :] = 30
    data[2, :] = 40
    data[3, :] = 50
    data[4, :] = np.nan
    data[5, :] = np.nan
    data[6, :] = np.nan
    data[7, :] = np.nan
    data[0, 0] = np.nan
    data[1, 1] = np.nan
    data[4, 2] = 7
    data[5, 3] = 8
    data[8, 0] = np.nan
    data[9, 1] = np.nan
    df         = pd.DataFrame(data.T)

    with np.errstate(divide='ignore'):
        nacorr   = core.naCorrelation(pd.isna(df), 0.1)
        colpairs = np.where(np.triu(nacorr, k=1))
        colpairs = np.vstack(colpairs).T
        pairwise = core.pairwiseRedundantColumns(df, colpairs, 0.2)
        matrix   = core.matrixRedundantColumns(  df, 0.2, 0.1)

    pairwise = np.array(sorted(pairwise))
    matrix   = np.array(sorted(matrix))
    assert np.all(pairwise == matrix)


def test_binariseCategorical():

    nvals = 100
    ncols = 100
    nrows = 1000
    data = np.random.randint(1, nvals + 1, (nrows, ncols))
    cols = {str(i + 1) : data[:, i] for i in range(ncols)}
    df   = pd.DataFrame(cols)

    bindata, uniq = core.binariseCategorical(df)

    assert sorted(uniq) == sorted(np.unique(data))

    for i, v in enumerate(uniq):
        exp = np.any(data == v, axis=1)
        assert np.all(bindata[:, i] == exp)

    data[data == 5] = 6
    data[:10, 0] = 5
    cols = {str(i + 1) : data[:, i] for i in range(ncols)}
    df   = pd.DataFrame(cols)

    bindata, uniq = core.binariseCategorical(df, minpres=11)

    assert sorted(uniq) == sorted(set(range(1, nvals + 1)).difference([5]))

    for i, v in enumerate(uniq):
        exp = np.any(data == v, axis=1)
        assert np.all(bindata[:, i] == exp)


def test_binariseCategorical_missing():

    data = np.full((50, 5), np.nan)

    for i in range(data.shape[1]):
        namask = np.random.random(data.shape[0]) < 0.1
        data[~namask, i] = np.random.randint(1, 10, (~namask).sum())

    expuniq = list(sorted(np.unique(data[~np.isnan(data)])))

    expdata = np.zeros((data.shape[0], len(expuniq)))

    for i, v in enumerate(expuniq):
        expdata[:, i] = np.any(data == v, axis=1)

    cols = {str(i + 1) : data[:, i] for i in range(data.shape[1])}
    df   = pd.DataFrame(cols)

    gotdata, gotuniq = core.binariseCategorical(df)
    assert np.all(gotuniq == expuniq)
    assert np.all(gotdata == expdata)


def test_binariseCategorical_take():

    nvals = 100
    ncols = 100
    nrows = 1000

    data = np.full((nrows, ncols), np.nan)
    aux  = np.full((nrows, ncols), np.nan)

    for i in range(data.shape[1]):
        namask = np.random.random(data.shape[0]) < 0.7
        data[~namask, i] = np.random.randint(1, nvals + 1, (~namask).sum())

    mask      = np.isfinite(data)
    aux[mask] = np.random.randint(1, 100, data.shape)[mask]

    expuniq = list(sorted(np.unique(data[~np.isnan(data)])))

    expdata = np.full((data.shape[0], len(expuniq)), np.nan)

    for i, v in enumerate(expuniq):
        rows, cols = np.where(data == v)
        for r in range(data.shape[0]):
            if r not in rows:
                continue
            c = cols[np.where(rows == r)[0][0]]
            expdata[r, i] = aux[r, c]

    cols = {str(i + 1) : data[:, i] for i in range(data.shape[1])}
    df   = pd.DataFrame(cols)

    cols = {str(i + 1) : aux[:, i] for i in range(aux.shape[1])}
    adf  = pd.DataFrame(cols)

    gotdata, gotuniq = core.binariseCategorical(df, take=adf)

    gotna = pd.isnull(gotdata)
    expna = pd.isnull(expdata)

    assert np.all(gotuniq         == expuniq)
    assert np.all(gotna           == expna)
    assert np.all(gotdata[~gotna] == expdata[~expna])


def test_binariseCategorical_take_error():

    data = pd.DataFrame(np.random.randint(1, 10, (100, 10)))
    take = pd.DataFrame(np.random.randint(1, 10, (100, 9)))

    with pytest.raises(ValueError):
        core.binariseCategorical(data, take=take)


def test_binariseCategorical_parallel():

    nrows = 10000
    ncols = 100
    nvals = 500
    data = np.random.randint(1, nvals + 1, (nrows, ncols))
    cols = {str(i + 1) : data[:, i] for i in range(ncols)}
    df   = pd.DataFrame(cols)

    seqbindata, sequniq = core.binariseCategorical(df)
    parbindata, paruniq = core.binariseCategorical(df, njobs=6)

    assert np.all(seqbindata == parbindata)
    assert np.all(sequniq == paruniq)

    aux  = np.random.randint(nvals, nvals + 500, (nrows, ncols))
    cols = {str(i + 1) : aux[:, i] for i in range(aux.shape[1])}
    adf  = pd.DataFrame(cols)

    seqbindata, sequniq = core.binariseCategorical(df, take=adf)
    parbindata, paruniq = core.binariseCategorical(df, take=adf, njobs=6)

    assert np.all(seqbindata == parbindata)
    assert np.all(sequniq == paruniq)


def test_binariseCategorical_datetime():
    nrows = 5
    ncols = 10
    nvals = 10
    data = np.random.randint(1, nvals + 1, (nrows, ncols))
    cols = {str(i + 1) : data[:, i] for i in range(ncols)}
    df   = pd.DataFrame(cols)

    aux  = np.random.randint(100, 1000000, (nrows, ncols),
                             dtype=np.uint64).astype('<M8[s]')
    cols = {str(i + 1) : aux[:, i] for i in range(ncols)}
    adf  = pd.DataFrame(cols)

    seqbindata, sequniq = core.binariseCategorical(df, take=adf)
    parbindata, paruniq = core.binariseCategorical(df, take=adf, njobs=6)

    expuniq = list(sorted(np.unique(data)))
    expdata = np.full((data.shape[0], len(expuniq)), np.nan, '<M8[s]')

    for i, v in enumerate(expuniq):
        rows, cols = np.where(data == v)
        for r in range(data.shape[0]):
            if r not in rows:
                continue
            c = cols[np.where(rows == r)[0][0]]
            expdata[r, i] = aux[r, c]

    expnan = np.isnan(expdata)

    assert np.all(sequniq              == expuniq)
    assert np.all(paruniq              == expuniq)
    assert np.all(np.isnan(seqbindata) == expnan)
    assert np.all(np.isnan(parbindata) == expnan)
    assert np.all(expdata[~expnan]     == seqbindata[~expnan])
    assert np.all(expdata[~expnan]     == parbindata[~expnan])


# https://git.fmrib.ox.ac.uk/fsl/funpack/-/merge_requests/81
def test_binariseCategorical_empty():

    data = np.full(10, np.nan, dtype=np.float32)
    data = pd.DataFrame(data)

    columns, values = core.binariseCategorical(data)

    assert columns.shape[1] == 0
    assert len(values)      == 0


def test_expandCompound():

    data = []

    for i in range(20):
        rlen = np.random.randint(1, 20)
        row = np.random.randint(1, 100, rlen)
        data.append(row)

    exp = np.full((20, max(map(len, data))), np.nan)

    for i in range(20):
        exp[i, :len(data[i])] = data[i]

    series = pd.Series(data)

    got = core.expandCompound(series)

    expna = np.isnan(exp)
    gotna = np.isnan(got)

    assert np.all(     expna  ==      gotna)
    assert np.all(exp[~expna] == got[~gotna])
